"""
FiftyOne dataset types.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import eta.core.utils as etau


class Dataset(object):
    """Base type for datasets."""

    def get_dataset_importer_cls(self):
        """Returns the :class:`fiftyone.utils.data.importers.DatasetImporter`
        class for importing datasets of this type from disk.

        Returns:
            a :class:`fiftyone.utils.data.importers.DatasetImporter` class
        """
        raise TypeError(
            "Dataset type '%s' does not support imports"
            % etau.get_class_name(self)
        )

    def get_dataset_exporter_cls(self):
        """Returns the :class:`fiftyone.utils.data.exporters.DatasetExporter`
        class for exporting datasets of this type to disk.

        Returns:
            a :class:`fiftyone.utils.data.exporters.DatasetExporter` class
        """
        raise TypeError(
            "Dataset type '%s' does not support exports"
            % etau.get_class_name(self)
        )


class UnlabeledDataset(Dataset):
    """Base type for datasets that represent an unlabeled collection of data
    samples.
    """

    pass


class UnlabeledImageDataset(UnlabeledDataset):
    """Base type for datasets that represent an unlabeled collection of images."""

    def get_dataset_importer_cls(self):
        """Returns the
        :class:`fiftyone.utils.data.importers.UnlabeledImageDatasetImporter`
        class for importing datasets of this type from disk.

        Returns:
            a :class:`fiftyone.utils.data.importers.UnlabeledImageDatasetImporter`
            class
        """
        return super().get_dataset_importer_cls()

    def get_dataset_exporter_cls(self):
        """Returns the
        :class:`fiftyone.utils.data.exporters.UnlabeledImageDatasetExporter`
        class for exporting datasets of this type to disk.

        Returns:
            a :class:`fiftyone.utils.data.exporters.UnlabeledImageDatasetExporter`
            class
        """
        return super().get_dataset_exporter_cls()


class UnlabeledVideoDataset(UnlabeledDataset):
    """Base type for datasets that represent an unlabeled collection of videos."""

    def get_dataset_importer_cls(self):
        """Returns the
        :class:`fiftyone.utils.data.importers.UnlabeledVideoDatasetImporter`
        class for importing datasets of this type from disk.

        Returns:
            a :class:`fiftyone.utils.data.importers.UnlabeledVideoDatasetImporter`
            class
        """
        return super().get_dataset_importer_cls()

    def get_dataset_exporter_cls(self):
        """Returns the
        :class:`fiftyone.utils.data.exporters.UnlabeledVideoDatasetExporter`
        class for exporting datasets of this type to disk.

        Returns:
            a :class:`fiftyone.utils.data.exporters.UnlabeledVideoDatasetExporter`
            class
        """
        return super().get_dataset_exporter_cls()


class LabeledDataset(Dataset):
    """Base type for datasets that represent a collection of data samples and
    their associated labels.
    """

    pass


class LabeledImageDataset(LabeledDataset):
    """Base type for datasets that represent a collection of images and their
    associated labels.
    """

    def get_dataset_importer_cls(self):
        """Returns the
        :class:`fiftyone.utils.data.importers.LabeledImageDatasetImporter`
        class for importing datasets of this type from disk.

        Returns:
            a :class:`fiftyone.utils.data.importers.LabeledImageDatasetImporter`
            class
        """
        return super().get_dataset_importer_cls()

    def get_dataset_exporter_cls(self):
        """Returns the
        :class:`fiftyone.utils.data.exporters.LabeledImageDatasetExporter`
        class for exporting datasets of this type to disk.

        Returns:
            a :class:`fiftyone.utils.data.exporters.LabeledImageDatasetExporter`
            class
        """
        return super().get_dataset_exporter_cls()


class LabeledVideoDataset(LabeledDataset):
    """Base type for datasets that represent a collection of videos and their
    associated labels.
    """

    def get_dataset_importer_cls(self):
        """Returns the
        :class:`fiftyone.utils.data.importers.LabeledVideoDatasetImporter`
        class for importing datasets of this type from disk.

        Returns:
            a :class:`fiftyone.utils.data.importers.LabeledVideoDatasetImporter`
            class
        """
        return super().get_dataset_importer_cls()

    def get_dataset_exporter_cls(self):
        """Returns the
        :class:`fiftyone.utils.data.exporters.LabeledVideoDatasetExporter`
        class for exporting datasets of this type to disk.

        Returns:
            a :class:`fiftyone.utils.data.exporters.LabeledVideoDatasetExporter`
            class
        """
        return super().get_dataset_exporter_cls()


class ImageClassificationDataset(LabeledImageDataset):
    """Base type for datasets that represent a collection of images and a set
    of associated classification labels.
    """

    pass


class VideoClassificationDataset(LabeledVideoDataset):
    """Base type for datasets that represent a collection of videos and a set
    of associated classification labels.
    """

    pass


class ImageDetectionDataset(LabeledImageDataset):
    """Base type for datasets that represent a collection of images and a set
    of associated detections.
    """

    pass


class VideoDetectionDataset(LabeledVideoDataset):
    """Base type for datasets that represent a collection of videos and a set
    of associated video detections.
    """

    pass


class ImageSegmentationDataset(LabeledImageDataset):
    """Base type for datasets that represent a collection of images and a set
    of associated semantic segmentations.
    """

    pass


class ImageLabelsDataset(LabeledImageDataset):
    """Base type for datasets that represent a collection of images and a set
    of associated multitask predictions.
    """

    pass


class VideoLabelsDataset(LabeledVideoDataset):
    """Base type for datasets that represent a collection of videos and a set
    of associated multitask predictions.
    """

    pass


class GroupDataset(Dataset):
    """Base type for datasets that contain grouped samples of any type(s)."""

    def get_dataset_importer_cls(self):
        """Returns the
        :class:`fiftyone.utils.data.importers.GroupDatasetImporter` class for
        importing datasets of this type from disk.

        Returns:
            a :class:`fiftyone.utils.data.importers.GroupDatasetImporter` class
        """
        return super().get_dataset_importer_cls()

    def get_dataset_exporter_cls(self):
        """Returns the
        :class:`fiftyone.utils.data.exporters.GroupDatasetExporter` class for
        exporting datasets of this type to disk.

        Returns:
            a :class:`fiftyone.utils.data.exporters.GroupDatasetExporter` class
        """
        return super().get_dataset_exporter_cls()


class ImageDirectory(UnlabeledImageDataset):
    """A directory of images.

    See :ref:`this page <ImageDirectory-import>` for importing datasets of this
    type, and see :ref:`this page <ImageDirectory-export>` for exporting
    datasets of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.data as foud

        return foud.ImageDirectoryImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.data as foud

        return foud.ImageDirectoryExporter


class VideoDirectory(UnlabeledImageDataset):
    """A directory of videos.

    See :ref:`this page <VideoDirectory-import>` for importing datasets of this
    type, and see :ref:`this page <VideoDirectory-export>` for exporting
    datasets of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.data as foud

        return foud.VideoDirectoryImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.data as foud

        return foud.VideoDirectoryExporter


class FiftyOneImageClassificationDataset(ImageClassificationDataset):
    """A labeled dataset consisting of images and their associated
    classification labels stored in a simple JSON format.

    See :ref:`this page <FiftyOneImageClassificationDataset-import>` for
    importing datasets of this type, and see
    :ref:`this page <FiftyOneImageClassificationDataset-export>` for exporting
    datasets of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.data as foud

        return foud.FiftyOneImageClassificationDatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.data as foud

        return foud.FiftyOneImageClassificationDatasetExporter


class ImageClassificationDirectoryTree(ImageClassificationDataset):
    """A directory tree whose subfolders define an image classification
    dataset.

    See :ref:`this page <ImageClassificationDirectoryTree-import>` for
    importing datasets of this type, and see
    :ref:`this page <ImageClassificationDirectoryTree-export>` for exporting
    datasets of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.data as foud

        return foud.ImageClassificationDirectoryTreeImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.data as foud

        return foud.ImageClassificationDirectoryTreeExporter


class VideoClassificationDirectoryTree(VideoClassificationDataset):
    """A directory tree whose subfolders define a video classification dataset.

    See :ref:`this page <VideoClassificationDirectoryTree-import>` for
    importing datasets of this type, and see
    :ref:`this page <VideoClassificationDirectoryTree-export>` for exporting
    datasets of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.data as foud

        return foud.VideoClassificationDirectoryTreeImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.data as foud

        return foud.VideoClassificationDirectoryTreeExporter


class TFImageClassificationDataset(ImageClassificationDataset):
    """A labeled dataset consisting of images and their associated
    classification labels stored as
    `TFRecords <https://www.tensorflow.org/tutorials/load_data/tfrecord>`_.

    See :ref:`this page <TFImageClassificationDataset-import>` for importing
    datasets of this type, and see
    :ref:`this page <TFImageClassificationDataset-export>` for exporting
    datasets of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.tf as fout

        return fout.TFImageClassificationDatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.tf as fout

        return fout.TFImageClassificationDatasetExporter


class FiftyOneImageDetectionDataset(ImageDetectionDataset):
    """A labeled dataset consisting of images and their associated object
    detections stored in a simple JSON format.

    See :ref:`this page <FiftyOneImageDetectionDataset-import>` for importing
    datasets of this type, and see
    :ref:`this page <FiftyOneImageDetectionDataset-export>` for exporting
    datasets of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.data as foud

        return foud.FiftyOneImageDetectionDatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.data as foud

        return foud.FiftyOneImageDetectionDatasetExporter


class FiftyOneTemporalDetectionDataset(VideoDetectionDataset):
    """A labeled dataset consisting of videos and their associated temporal
    detections stored in a simple JSON format.

    See :ref:`this page <FiftyOneTemporalDetectionDataset-import>` for
    importing datasets of this type, and see
    :ref:`this page <FiftyOneTemporalDetectionDataset-export>` for exporting
    datasets of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.data as foud

        return foud.FiftyOneTemporalDetectionDatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.data as foud

        return foud.FiftyOneTemporalDetectionDatasetExporter


class COCODetectionDataset(ImageDetectionDataset):
    """A labeled dataset consisting of images and their associated object
    detections saved in
    `COCO Object Detection Format <https://cocodataset.org/#format-data>`_.

    See :ref:`this page <COCODetectionDataset-import>` for importing datasets
    of this type, and see :ref:`this page <COCODetectionDataset-export>` for
    exporting datasets of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.coco as fouc

        return fouc.COCODetectionDatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.coco as fouc

        return fouc.COCODetectionDatasetExporter


class VOCDetectionDataset(ImageDetectionDataset):
    """A labeled dataset consisting of images and their associated object
    detections saved in `VOC format <http://host.robots.ox.ac.uk/pascal/VOC>`_.

    See :ref:`this page <VOCDetectionDataset-import>` for importing datasets of
    this type, and see :ref:`this page <VOCDetectionDataset-export>` for
    exporting datasets of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.voc as fouv

        return fouv.VOCDetectionDatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.voc as fouv

        return fouv.VOCDetectionDatasetExporter


class KITTIDetectionDataset(ImageDetectionDataset):
    """A labeled dataset consisting of images and their associated object
    detections saved in
    `KITTI format <http://www.cvlibs.net/datasets/kitti/eval_object.php>`_.

    See :ref:`this page <KITTIDetectionDataset-import>` for importing datasets
    of this type, and see :ref:`this page <KITTIDetectionDataset-export>` for
    exporting datasets of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.kitti as fouk

        return fouk.KITTIDetectionDatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.kitti as fouk

        return fouk.KITTIDetectionDatasetExporter


class OpenImagesV6Dataset(ImageDetectionDataset):
    """A labeled dataset consisting of images and their associated annotations
    saved in
    `Open Images format <https://storage.googleapis.com/openimages/web/download.html>`_.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.openimages as fouo

        return fouo.OpenImagesV6DatasetImporter


class FIWDataset(Dataset):
    """A labeled dataset consisting of images and their associated annotations
    saved in
    `Families in the Wild format <https://github.com/visionjo/pykinship#db-contents-and-structure>`_.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.fiw as fouf

        return fouf.FIWDatasetImporter


class OpenLABELImageDataset(ImageLabelsDataset):
    def get_dataset_importer_cls(self):
        import fiftyone.utils.openlabel as fouo

        return fouo.OpenLABELImageDatasetImporter


class OpenLABELVideoDataset(VideoLabelsDataset):
    def get_dataset_importer_cls(self):
        import fiftyone.utils.openlabel as fouo

        return fouo.OpenLABELVideoDatasetImporter


class YOLOv4Dataset(ImageDetectionDataset):
    """A labeled dataset consisting of images and their associated object
    detections saved in `YOLOv4 format <https://github.com/AlexeyAB/darknet>`_.

    See :ref:`this page <YOLOv4Dataset-import>` for importing datasets of this
    type, and see :ref:`this page <YOLOv4Dataset-export>` for exporting datasets
    of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.yolo as fouy

        return fouy.YOLOv4DatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.yolo as fouy

        return fouy.YOLOv4DatasetExporter


class YOLOv5Dataset(ImageDetectionDataset):
    """A labeled dataset consisting of images and their associated object
    detections saved in
    `YOLOv5 format <https://github.com/ultralytics/yolov5>`_.

    See :ref:`this page <YOLOv5Dataset-import>` for importing datasets of this
    type, and see :ref:`this page <YOLOv5Dataset-export>` for exporting datasets
    of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.yolo as fouy

        return fouy.YOLOv5DatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.yolo as fouy

        return fouy.YOLOv5DatasetExporter


class TFObjectDetectionDataset(ImageDetectionDataset):
    """A labeled dataset consisting of images and their associated object
    detections stored as TFRecords in
    `TF Object Detection API format <https://github.com/tensorflow/models/blob/master/research/object_detection>`_.

    See :ref:`this page <TFObjectDetectionDataset-import>` for importing
    datasets of this type, and see
    :ref:`this page <TFObjectDetectionDataset-export>` for exporting datasets
    of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.tf as fout

        return fout.TFObjectDetectionDatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.tf as fout

        return fout.TFObjectDetectionDatasetExporter


class ImageSegmentationDirectory(ImageSegmentationDataset):
    """An labeled dataset consisting of images and their associated semantic
    segmentations stored as images on disk.

    See :ref:`this page <ImageSegmentationDirectory-import>` for importing
    datasets of this type, and see
    :ref:`this page <ImageSegmentationDirectory-export>` for exporting datasets
    of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.data as foud

        return foud.ImageSegmentationDirectoryImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.data as foud

        return foud.ImageSegmentationDirectoryExporter


class CVATImageDataset(ImageLabelsDataset):
    """A labeled dataset consisting of images and their associated labels
    stored in `CVAT image format <https://github.com/opencv/cvat>`_.

    See :ref:`this page <CVATImageDataset-import>` for importing datasets of
    this type, and see :ref:`this page <CVATImageDataset-export>` for exporting
    datasets of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.cvat as fouc

        return fouc.CVATImageDatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.cvat as fouc

        return fouc.CVATImageDatasetExporter


class CVATVideoDataset(VideoLabelsDataset):
    """A labeled dataset consisting of images and their associated object
    detections stored in `CVAT video format <https://github.com/opencv/cvat>`_.

    See :ref:`this page <CVATVideoDataset-import>` for importing datasets of
    this type, and see :ref:`this page <CVATVideoDataset-export>` for exporting
    datasets of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.cvat as fouc

        return fouc.CVATVideoDatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.cvat as fouc

        return fouc.CVATVideoDatasetExporter


class FiftyOneImageLabelsDataset(ImageLabelsDataset):
    """A labeled dataset consisting of images and their associated multitask
    predictions stored in
    `ETA ImageLabels format <https://github.com/voxel51/eta/blob/develop/docs/image_labels_guide.md>`_.

    See :ref:`this page <FiftyOneImageLabelsDataset-import>` for importing
    datasets of this type, and see
    :ref:`this page <FiftyOneImageLabelsDataset-export>` for exporting datasets
    of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.data as foud

        return foud.FiftyOneImageLabelsDatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.data as foud

        return foud.FiftyOneImageLabelsDatasetExporter


class FiftyOneVideoLabelsDataset(VideoLabelsDataset):
    """A labeled dataset consisting of videos and their associated labels
    stored in
    `ETA VideoLabels format <https://github.com/voxel51/eta/blob/develop/docs/video_labels_guide.md>`_.

    See :ref:`this page <FiftyOneVideoLabelsDataset-import>` for importing
    datasets of this type, and see
    :ref:`this page <FiftyOneVideoLabelsDataset-export>` for exporting datasets
    of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.data as foud

        return foud.FiftyOneVideoLabelsDatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.data as foud

        return foud.FiftyOneVideoLabelsDatasetExporter


class BDDDataset(ImageLabelsDataset):
    """A labeled dataset consisting of images and their associated multitask
    predictions saved in
    `Berkeley DeepDrive (BDD) format <https://bdd-data.berkeley.edu>`_.

    See :ref:`this page <BDDDataset-import>` for importing datasets of this
    type, and see :ref:`this page <BDDDataset-export>` for exporting datasets
    of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.bdd as foub

        return foub.BDDDatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.bdd as foub

        return foub.BDDDatasetExporter


class DICOMDataset(ImageLabelsDataset):
    """An image dataset whose image data and optional properties
    are stored in `DICOM format <https://en.wikipedia.org/wiki/DICOM>`_.

    See :ref:`this page <DICOMDataset-import>` for importing datasets of this
    type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.dicom as foud

        return foud.DICOMDatasetImporter


class ActivityNetDataset(FiftyOneTemporalDetectionDataset):
    """A video dataset composed of temporal activity detections from the
    `ActivityNet dataset <http://activity-net.org/download.html>`_.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.activitynet as foua

        return foua.ActivityNetDatasetImporter


class GeoJSONDataset(LabeledDataset):
    """An image or video dataset whose geolocation data and optional properties
    are stored in `GeoJSON format <https://en.wikipedia.org/wiki/GeoJSON>`_.

    See :ref:`this page <GeoJSONDataset-import>` for importing datasets of this
    type, and see :ref:`this page <GeoJSONDataset-export>` for exporting
    datasets of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.geojson as foug

        return foug.GeoJSONDatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.geojson as foug

        return foug.GeoJSONDatasetExporter


class GeoTIFFDataset(ImageLabelsDataset):
    """An image dataset whose image and geolocation data are stored in
    `GeoTIFF format <https://en.wikipedia.org/wiki/GeoTIFF>`_.

    See :ref:`this page <GeoTIFFDataset-import>` for importing datasets of this
    type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.geotiff as foug

        return foug.GeoTIFFDatasetImporter


class FiftyOneDataset(Dataset):
    """A disk representation of an entire
    :class:`fiftyone.core.dataset.Dataset` stored on disk in a serialized JSON
    format along with its source media.

    See :ref:`this page <FiftyOneDataset-import>` for importing datasets of
    this type, and see :ref:`this page <FiftyOneDataset-export>` for exporting
    datasets of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.data as foud

        return foud.FiftyOneDatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.data as foud

        return foud.FiftyOneDatasetExporter


class LegacyFiftyOneDataset(Dataset):
    """Legacy disk representation of an entire
    :class:`fiftyone.core.dataset.Dataset` stored on disk in a serialized JSON
    format along with its source media.

    Datasets of this type are read/written in the following format::

        <dataset_dir>/
            metadata.json
            samples.json
            data/
                <filename1>.<ext>
                <filename2>.<ext>
                ...
            annotations/
                <anno_key1>.json
                <anno_key2>.json
                ...
            brain/
                <brain_key1>.json
                <brain_key2>.json
                ...
            evaluations/
                <eval_key1>.json
                <eval_key2>.json
                ...

    where ``metadata.json`` is a JSON file containing metadata associated with
    the dataset, ``samples.json`` is a JSON file containing a serialized
    representation of the samples in the dataset, ``annotations/`` contains any
    serialized :class:`fiftyone.core.annotations.AnnotationResults`, ``brain/``
    contains any serialized :class:`fiftyone.core.brain.BrainResults`, and
    ``evaluations/`` contains any serialized
    :class:`fiftyone.core.evaluations.EvaluationResults`.

    Video datasets have an additional ``frames/`` directory that contains a
    serialized representation of the frame labels for each video in the
    dataset.

    .. note::

        See :class:`fiftyone.utils.data.importers.LegacyFiftyOneDatasetImporter` for
        parameters that can be passed to methods like
        :meth:`Dataset.from_dir() <fiftyone.core.dataset.Dataset.from_dir>` to
        customize the import of datasets of this type.

    .. note::

        See :class:`fiftyone.utils.data.exporters.LegacyFiftyOneDatasetExporter` for
        parameters that can be passed to methods like
        :meth:`SampleCollection.export() <fiftyone.core.collections.SampleCollection.export>`
        to customize the export of datasets of this type.
    """

    def get_dataset_importer_cls(self):
        import fiftyone.utils.data as foud

        return foud.LegacyFiftyOneDatasetImporter

    def get_dataset_exporter_cls(self):
        import fiftyone.utils.data as foud

        return foud.LegacyFiftyOneDatasetExporter



"""
Utilities for working with `YouTube <https://youtube.com>`.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import importlib
import itertools
import logging
import multiprocessing
import multiprocessing.dummy
import os
import pkg_resources

import numpy as np

import eta.core.utils as etau
import eta.core.video as etav

import fiftyone.core.utils as fou


def _ensure_pytube():
    fou.ensure_package("pytube>=11.0.2")
    if pkg_resources.get_distribution("pytube").version == "11.0.2":
        _patch_pytube_cypher()


pytube = fou.lazy_import("pytube", callback=_ensure_pytube)


logger = logging.getLogger(__name__)


def download_youtube_videos(
    urls,
    download_dir=None,
    video_paths=None,
    clip_segments=None,
    ext=".mp4",
    only_progressive=True,
    resolution="highest",
    max_videos=None,
    num_workers=None,
    skip_failures=True,
    quiet=False,
):
    """Downloads a list of YouTube videos.

    The ``urls`` argument accepts a list of YouTube "watch" URLs::

        urls = [
            "https://www.youtube.com/watch?v=-0URMJE8_PB",
            ...,
        ]

    Use either the ``download_dir`` or ``video_paths`` argument to specify
    where to download each video.

    You can use the optional ``clip_segments`` argument to specify a specific
    segment, in seconds, of each video to download::

        clip_segments = [
            (10, 25),
            (11.1, 20.2),
            None,               # entire video
            (None, 8.0),        # through beginning of video
            (8.0, None),        # through end of video
            ...
        ]

    You can also use the optional ``ext`` and ``resolution`` arguments to
    specify a deisred video codec and resolution to download, if possible.

    YouTube videos are regularly taken down. Therefore, this method provides an
    optional ``max_videos`` argument that you can use in conjunction with
    ``skip_failures=True`` and a large list of possibly non-existent videos in
    ``urls`` in cases where you need a certain number of videos to be
    successfully downloaded but are willing to tolerate failures.

    Args:
        urls: a list of YouTube URLs to download
        download_dir (None): a directory in which to store the downloaded
            videos
        video_paths (None): a list of paths to which to download the videos.
            When downloading entire videos, a stream matching the video format
            implied by each file's extension is downloaded, if available, or
            else the extension of the video path is **changed** to match the
            available stream's format
        clip_segments (None): a list of ``(first, last)`` tuples defining a
            specific segment of each video to download
        ext (".mp4"): a video format to download for each video, if possible.
            Only applicable when a ``download_dir`` is used. This format will
            be respected if such a stream exists, otherwise the format of the
            best available stream is used. Set this value to ``None`` if you
            want to download the stream with the best match for ``resolution``
            and ``progressive`` regardless of format
        only_progressive (True): whether to only download progressive streams,
            if possible. Progressive streams contain both audio and video
            tracks and are typically only available at <= 720p resolution
        resolution ("highest"): a desired stream resolution to download. This
            filter is applied after respecting the desired video format and
            ``only_progressive`` restriction, if applicable. The supported
            values are:

            -   ``"highest"`` (default): download the highest resolution stream
            -   ``"lowest"``:  download the lowest resolution stream
            -   A target resolution like ``"1080p"``. In this case, the stream
                whose resolution is closest to this target value is downloaded
        max_videos (None): the maximum number of videos to successfully
            download. By default, all videos are downloaded
        num_workers (None): the number of threads or processes to use when
            downloading videos. By default, ``multiprocessing.cpu_count()`` is
            used
        skip_failures (True): whether to gracefully continue without raising
            an error if a video cannot be downloaded
        quiet (False): whether to suppress logging, except for a progress bar

    Returns:
        a tuple of

        -   **downloaded**: a dict mapping integer indexes into ``urls`` to
            paths of successfully downloaded videos
        -   **errors**: a dict mapping integer indexes into ``urls`` to error
            messages for videos that were attempted to be downloaded, but
            failed
    """
    use_threads = clip_segments is None
    num_workers = _parse_num_workers(num_workers, use_threads=use_threads)

    if max_videos is None:
        max_videos = len(urls)

    with etau.TempDir() as tmp_dir:
        tasks = _build_tasks_list(
            urls,
            download_dir,
            video_paths,
            clip_segments,
            tmp_dir,
            ext,
            only_progressive,
            resolution,
        )

        if num_workers <= 1:
            downloaded, errors = _download(
                tasks, max_videos, skip_failures, quiet
            )
        else:
            downloaded, errors = _download_multi(
                tasks,
                max_videos,
                skip_failures,
                quiet,
                num_workers,
                use_threads,
            )

    return downloaded, errors


def _parse_num_workers(num_workers, use_threads=False):
    if num_workers is None:
        if os.name == "nt" and not use_threads:
            # Multiprocessing on Windows is bad news
            return 1

        return multiprocessing.cpu_count()

    return num_workers


def _build_tasks_list(
    urls,
    download_dir,
    video_paths,
    clip_segments,
    tmp_dir,
    ext,
    only_progressive,
    resolution,
):
    if video_paths is None and download_dir is None:
        raise ValueError("Either `download_dir` or `video_paths` are required")

    if video_paths is not None and download_dir is not None:
        raise ValueError(
            "Only one of `download_dir` or `video_paths` can be provided"
        )

    if download_dir is None:
        ext = None

    if not etau.is_str(resolution) or (
        resolution not in ("highest", "lowest")
        and not resolution.endswith("p")
    ):
        raise ValueError(
            "Invalid resolution=%s. The supported values are 'highest', "
            "'lowest', and strings like '1080p', '720p', ..." % resolution
        )

    if resolution.endswith("p"):
        resolution = int(resolution[:-1])

    if video_paths is None:
        video_paths = itertools.repeat(None)

    if clip_segments is None:
        clip_segments = itertools.repeat(clip_segments)
    else:
        clip_segments = list(clip_segments)

        # Replace all variations of full video downloads with `None` so that
        # the most optimized download logic is used
        for idx, clip_segment in enumerate(clip_segments):
            if (
                clip_segment is not None
                and (clip_segment[0] is None or clip_segment[0] <= 0)
                and clip_segment[1] is None
            ):
                clip_segments[idx] = None

    return list(
        zip(
            range(len(urls)),
            urls,
            itertools.repeat(download_dir),
            video_paths,
            clip_segments,
            itertools.repeat(tmp_dir),
            itertools.repeat(ext),
            itertools.repeat(only_progressive),
            itertools.repeat(resolution),
        )
    )


def _download(tasks, max_videos, skip_failures, quiet):
    downloaded = {}
    errors = {}

    with fou.ProgressBar(total=max_videos, iters_str="videos") as pb:
        for task in tasks:
            idx, url, video_path, error, warnings = _do_download(task)

            if warnings and not quiet:
                for msg in warnings:
                    logger.warning(msg)

            if error:
                msg = "Failed to download video '%s': %s" % (url, error)
                if skip_failures:
                    if not quiet:
                        logger.warning(msg)
                else:
                    raise ValueError(msg)

                errors[idx] = error
                pb.draw()
            else:
                pb.update()
                downloaded[idx] = video_path
                if len(downloaded) >= max_videos:
                    return downloaded, errors

    return downloaded, errors


def _download_multi(
    tasks, max_videos, skip_failures, quiet, num_workers, use_threads
):
    downloaded = {}
    errors = {}

    with fou.ProgressBar(total=max_videos, iters_str="videos") as pb:
        if use_threads:
            pool_cls = multiprocessing.dummy.Pool
        else:
            pool_cls = multiprocessing.Pool

        with pool_cls(num_workers) as pool:
            for idx, url, video_path, error, warnings in pool.imap_unordered(
                _do_download, tasks
            ):
                if warnings and not quiet:
                    for msg in warnings:
                        logger.warning(msg)

                if error:
                    msg = "Failed to download video '%s': %s" % (url, error)
                    if skip_failures:
                        if not quiet:
                            logger.warning(msg)
                    else:
                        raise ValueError(msg)

                    errors[idx] = error
                    pb.draw()
                else:
                    pb.update()
                    downloaded[idx] = video_path
                    if len(downloaded) >= max_videos:
                        return downloaded, errors

    return downloaded, errors


def _do_download(task):
    (
        idx,
        url,
        download_dir,
        video_path,
        clip_segment,
        tmp_dir,
        ext,
        only_progressive,
        resolution,
    ) = task

    error = None
    warnings = []

    try:
        pytube_video = pytube.YouTube(url)
        _validate_video(pytube_video)

        if video_path is not None and ext is None:
            ext = os.path.splitext(video_path)[1]

        stream = _get_stream(pytube_video, ext, only_progressive, resolution)

        if video_path is None:
            filename = stream.default_filename
            if ext is not None:
                filename = os.path.splitext(filename)[0] + ext

            video_path = os.path.join(download_dir, filename)

        root, ext = os.path.splitext(video_path)
        stream_ext = os.path.splitext(stream.default_filename)[1]
        if ext != stream_ext:
            warnings.append(
                (
                    "Unable to find a '%s' stream for '%s'; downloading '%s' "
                    "instead"
                )
                % (ext, url, stream_ext)
            )

            video_path = root + stream_ext

        if only_progressive and not stream.is_progressive:
            warnings.append(
                (
                    "Unable to find a progressive stream for '%s'; "
                    "downloading a non-progressive stream instead"
                )
                % url
            )

        if etau.is_numeric(resolution):
            target_res = "%dp" % resolution
            stream_res = stream.resolution
            if target_res != stream_res:
                warnings.append(
                    (
                        "Unable to find a '%s' stream for '%s'; downloading a "
                        "'%s' stream instead"
                    )
                    % (target_res, url, stream_res)
                )

        # Download to a temporary location first and then move to `video_path`
        # so that only successful downloads end up at their final destination
        tmp_path = os.path.join(tmp_dir, os.path.basename(video_path))

        if clip_segment is None:
            _download_video(stream, tmp_path)
        else:
            _download_clip(stream, clip_segment, tmp_path)

        etau.move_file(tmp_path, video_path)
    except Exception as e:
        video_path = None
        if isinstance(e, pytube.exceptions.PytubeError):
            error = type(e)
        else:
            error = str(e)

    return idx, url, video_path, error, warnings


def _validate_video(pytube_video):
    status, messages = pytube.extract.playability_status(
        pytube_video.watch_html
    )

    if status is None:
        return None

    if not etau.is_container(messages):
        error = messages
    elif messages:
        error = messages[0]
    else:
        error = status

    raise ValueError(error)


def _get_stream(pytube_video, ext, only_progressive, resolution):
    if ext is not None:
        ext = ext[1:]  # remove "."

    if only_progressive:
        progressive = True
    else:
        progressive = None

    while True:
        streams = pytube_video.streams.filter(
            type="video", progressive=progressive, file_extension=ext
        )

        if streams:
            if etau.is_numeric(resolution):
                all_res = [int(s.resolution[:-1]) for s in streams]
                idx = _find_nearest(all_res, resolution)
                return streams[idx]

            if resolution == "lowest":
                return streams.order_by("resolution").first()

            return streams.order_by("resolution").desc().first()

        if progressive:
            progressive = None
        elif ext is not None:
            if only_progressive and progressive is None:
                progressive = True

            ext = None
        else:
            raise ValueError("No video streams found")


def _find_nearest(array, target):
    return np.argmin(np.abs(np.asarray(array) - target))


def _download_video(stream, video_path):
    outdir, filename = os.path.split(video_path)
    stream.download(
        output_path=outdir,
        filename=filename,
        skip_existing=False,
        max_retries=3,
    )


def _download_clip(stream, clip_segment, video_path):
    if clip_segment is None:
        clip_segment = (None, None)

    start_time, end_time = clip_segment

    if start_time is None:
        start_time = 0

    if end_time is not None:
        duration = end_time - start_time
    else:
        duration = None

    # @todo consider using `fast=True` here, or even using an nearest-keyframe
    # approach to further optimize the clip extraction
    etav.extract_clip(
        stream.url, video_path, start_time=start_time, duration=duration
    )


def _patch_pytube_cypher():
    filepath = os.path.normpath(
        os.path.join(
            os.path.dirname(importlib.util.find_spec("pytube").origin),
            "cipher.py",
        )
    )

    find = """
    function_patterns = [
        # https://github.com/ytdl-org/youtube-dl/issues/29326#issuecomment-865985377
        # a.C&&(b=a.get("n"))&&(b=Dea(b),a.set("n",b))}};
        # In above case, `Dea` is the relevant function name
        r'a\.[A-Z]&&\(b=a\.get\("n"\)\)&&\(b=([^(]+)\(b\)',
    ]
    logger.debug('Finding throttling function name')
    for pattern in function_patterns:
        regex = re.compile(pattern)
        function_match = regex.search(js)
        if function_match:
            logger.debug("finished regex search, matched: %s", pattern)
            return function_match.group(1)"""

    replace = """
    # Patched by FiftyOne: https://github.com/voxel51/fiftyone
    # PR: https://github.com/pytube/pytube/pull/1222
    function_patterns = [
        # https://github.com/ytdl-org/youtube-dl/issues/29326#issuecomment-865985377
        # https://github.com/yt-dlp/yt-dlp/commit/48416bc4a8f1d5ff07d5977659cb8ece7640dcd8
        # var Bpa = [iha];
        # ...
        # a.C && (b = a.get("n")) && (b = Bpa[0](b), a.set("n", b),
        # Bpa.length || iha("")) }};
        # In the above case, `iha` is the relevant function name
        r'a\.[a-zA-Z]\s*&&\s*\([a-z]\s*=\s*a\.get\("n"\)\)\s*&&\s*'
        r'\([a-z]\s*=\s*([a-zA-Z0-9$]{3})(\[\d+\])?\([a-z]\)',
    ]
    logger.debug('Finding throttling function name')
    for pattern in function_patterns:
        regex = re.compile(pattern)
        function_match = regex.search(js)
        if function_match:
            logger.debug("finished regex search, matched: %s", pattern)
            if len(function_match.groups()) == 1:
                return function_match.group(1)
            idx = function_match.group(2)
            if idx:
                idx = idx.strip("[]")
                array = re.search(
                    r'var {nfunc}\s*=\s*(\[.+?\]);'.format(
                        nfunc=function_match.group(1)),
                    js
                )
                if array:
                    array = array.group(1).strip("[]").split(",")
                    array = [x.strip() for x in array]
                    return array[int(idx)]"""

    try:
        with open(filepath, "r") as f:
            code = f.read()

        if find in code:
            logger.debug("Patching '%s'", filepath)
            fixed = code.replace(find, replace)
            with open(filepath, "w") as f:
                f.write(fixed)
        elif replace in code:
            logger.debug("Already patched '%s'", filepath)
        else:
            logger.debug("Unable to patch '%s'", filepath)
    except Exception as e:
        logger.debug(e)
        logger.debug("Unable to patch '%s'", filepath)



"""
GeoTIFF utilities.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import os

import eta.core.image as etai
import eta.core.utils as etau

import fiftyone.core.labels as fol
import fiftyone.core.metadata as fom
import fiftyone.core.utils as fou
import fiftyone.utils.data as foud

pyproj = fou.lazy_import("pyproj")
rasterio = fou.lazy_import(
    "rasterio", callback=lambda: fou.ensure_package("rasterio")
)


def get_geolocation(image_path):
    """Retrieves the geolocation information from the given GeoTIFF image.

    The returned :class:`fiftyone.core.labels.GeoLocation` will contain the
    lon/lat coordinates of the center of the image in its ``point`` attribute
    and the coordinates of its corners (clockwise, starting from the top-left)
    in its ``polygon`` attribute.

    Args:
        image_path: the path to the GeoTIFF image

    Returns:
        a :class:`fiftyone.core.labels.GeoLocation`
    """
    with open(image_path, "rb") as f:
        with rasterio.open(f, "r") as image:
            center = image.transform * (0.5 * image.width, 0.5 * image.height)

            proj = pyproj.Proj(image.crs)
            cp = proj(*center, inverse=True)
            tl = proj(image.bounds.left, image.bounds.top, inverse=True)
            tr = proj(image.bounds.right, image.bounds.top, inverse=True)
            br = proj(image.bounds.right, image.bounds.bottom, inverse=True)
            bl = proj(image.bounds.left, image.bounds.bottom, inverse=True)

    return fol.GeoLocation(point=cp, polygon=[[tl, tr, br, bl, tl]])


class GeoTIFFDatasetImporter(
    foud.LabeledImageDatasetImporter, foud.ImportPathsMixin
):
    """Importer for a directory of GeoTIFF images with geolocation data.

    See :ref:`this page <GeoTIFFDataset-import>` for format details.

    Args:
        dataset_dir (None): the dataset directory. If omitted, ``image_path``
            must be provided
        image_path (None): an optional parameter that enables explicit control
            over the location of the GeoTIFF images. Can be any of the
            following:

            -   a list of paths to GeoTIFF images. In this case,
                ``dataset_dir`` has no effect
            -   a glob pattern like ``"*.tif"`` specifying the location of the
                images in ``dataset_dir``
            -   an absolute glob pattern of GeoTIFF images. In this case,
                ``dataset_dir`` has no effect
        recursive (True): whether to recursively traverse subdirectories. Not
            applicable when ``image_path`` is provided
        compute_metadata (False): whether to produce
            :class:`fiftyone.core.metadata.ImageMetadata` instances for each
            image when importing
        shuffle (False): whether to randomly shuffle the order in which the
            samples are imported
        seed (None): a random seed to use when shuffling
        max_samples (None): a maximum number of samples to import. By default,
            all samples are imported
    """

    def __init__(
        self,
        dataset_dir=None,
        image_path=None,
        recursive=True,
        compute_metadata=False,
        shuffle=False,
        seed=None,
        max_samples=None,
    ):
        if dataset_dir is None and image_path is None:
            raise ValueError(
                "Either `dataset_dir` or `image_path` must be provided"
            )

        if not etau.is_container(image_path):
            image_path = self._parse_labels_path(
                dataset_dir=dataset_dir, labels_path=image_path
            )

        super().__init__(
            dataset_dir=dataset_dir,
            shuffle=shuffle,
            seed=seed,
            max_samples=max_samples,
        )

        self.image_path = image_path
        self.recursive = recursive
        self.compute_metadata = compute_metadata

        self._filepaths = None
        self._iter_filepaths = None
        self._num_samples = None

    def __iter__(self):
        self._iter_filepaths = iter(self._filepaths)
        return self

    def __len__(self):
        return self._num_samples

    def __next__(self):
        image_path = next(self._iter_filepaths)

        if self.compute_metadata:
            image_metadata = fom.ImageMetadata.build_for(image_path)
        else:
            image_metadata = None

        label = get_geolocation(image_path)

        return image_path, image_metadata, label

    @property
    def has_image_metadata(self):
        return self.compute_metadata

    @property
    def has_dataset_info(self):
        return False

    @property
    def label_cls(self):
        return fol.GeoLocation

    def setup(self):
        if self.image_path is not None:
            path = self.image_path

            if etau.is_str(path):
                if not os.path.isabs(path) and self.dataset_dir:
                    path = os.path.join(self.dataset_dir, path)

                # Glob pattern of images
                filepaths = etau.get_glob_matches(path)
            else:
                # List of images
                filepaths = list(path)
        else:
            # Directory of images
            filepaths = etau.list_files(
                self.dataset_dir, abs_paths=True, recursive=self.recursive
            )
            filepaths = [p for p in filepaths if etai.is_image_mime_type(p)]

        filepaths = self._preprocess_list(filepaths)

        self._filepaths = filepaths
        self._num_samples = len(filepaths)



"""
Utilities for working with the
`HMDB51 dataset <https://serre-lab.clps.brown.edu/resource/hmdb-a-large-human-motion-database>`_.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging
import os

import eta.core.utils as etau
import eta.core.web as etaw

import fiftyone.core.utils as fou


logger = logging.getLogger(__name__)


def download_hmdb51_dataset(
    dataset_dir, scratch_dir=None, fold=1, cleanup=True
):
    """Downloads and extracts the HMDB51 dataset.

    Any existing files are not re-downloaded.

    Args:
        dataset_dir: the directory to output the final dataset
        scratch_dir (None): a scratch directory to use to store temporary files
        fold (1): the test/train fold to use to arrange the files on disk. The
            supported values are ``(1, 2, 3)``
        cleanup (True): whether to cleanup the scratch directory after
            extraction
    """
    if scratch_dir is None:
        scratch_dir = os.path.join(dataset_dir, "scratch")

    if fold not in (1, 2, 3):
        raise ValueError("fold must be (1, 2, 3); found %s" % fold)

    # Download dataset
    videos_dir = _download_videos(scratch_dir)
    splits_dir = _download_splits(scratch_dir)

    # Parse splits info
    split_end = "_test_split%d.txt" % fold
    split_map = {
        os.path.basename(path)[: -len(split_end)]: path
        for path in etau.list_files(splits_dir, abs_paths=True)
        if path.endswith(split_end)
    }

    # Reorganize files into splits
    logger.info("Reorganizing videos into splits via fold %d...", fold)
    with fou.ProgressBar() as pb:
        for label, split_path in pb(split_map.items()):
            train_filenames, test_filenames = _load_split_info(split_path)

            # Train split
            for filename in train_filenames:
                inpath = os.path.join(videos_dir, label, filename)
                outpath = os.path.join(dataset_dir, "train", label, filename)
                etau.move_file(inpath, outpath)

            # Test split
            for filename in test_filenames:
                inpath = os.path.join(videos_dir, label, filename)
                outpath = os.path.join(dataset_dir, "test", label, filename)
                etau.move_file(inpath, outpath)

            # Other
            indir = os.path.join(videos_dir, label)
            outdir = os.path.join(dataset_dir, "other", label)
            etau.move_dir(indir, outdir)

    if cleanup:
        etau.delete_dir(scratch_dir)


_VIDEOS_DOWNLOAD_LINK = (
    "http://serre-lab.clps.brown.edu/wp-content/uploads/2013/10/hmdb51_org.rar"
)

_SPLITS_DOWNLOAD_LINK = (
    "http://serre-lab.clps.brown.edu/wp-content/uploads/2013/10/"
    "test_train_splits.rar"
)


def _download_videos(scratch_dir):
    rar_path = os.path.join(scratch_dir, "hmdb51_org.rar")
    videos_dir = os.path.join(scratch_dir, "videos")

    if not os.path.exists(rar_path):
        logger.info("Downloading dataset to '%s'", rar_path)
        etaw.download_file(_VIDEOS_DOWNLOAD_LINK, path=rar_path)
    else:
        logger.info("File '%s' already exists", rar_path)

    logger.info("Unpacking videos...")
    etau.extract_rar(rar_path, outdir=videos_dir, delete_rar=False)
    with fou.ProgressBar() as pb:
        for class_rar_path in pb(etau.list_files(videos_dir, abs_paths=True)):
            etau.extract_rar(
                class_rar_path, outdir=videos_dir, delete_rar=False
            )

    return videos_dir


def _download_splits(scratch_dir):
    rar_path = os.path.join(scratch_dir, "test_train_splits.rar")
    splits_dir = os.path.join(scratch_dir, "testTrainMulti_7030_splits")

    if not os.path.exists(rar_path):
        logger.info("Downloading split info to '%s'", rar_path)
        etaw.download_file(_SPLITS_DOWNLOAD_LINK, path=rar_path)
    else:
        logger.info("File '%s' already exists", rar_path)

    logger.info("Unpacking split info...")
    etau.extract_rar(rar_path, outdir=scratch_dir, delete_rar=False)

    return splits_dir


def _load_split_info(split_path):
    train = []
    test = []
    with open(split_path, "r") as f:
        for line in f.readlines():
            filename, split = line.strip().rsplit(maxsplit=1)
            if split == "1":
                train.append(filename)
            elif split == "2":
                test.append(filename)

    return train, test



"""
FiftyOne v0.17.0 revision.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""


def up(db, dataset_name):
    match_d = {"name": dataset_name}
    dataset_dict = db.datasets.find_one(match_d)

    if "group_field" not in dataset_dict:
        dataset_dict["group_field"] = None

    if "group_media_types" not in dataset_dict:
        dataset_dict["group_media_types"] = {}

    if "default_group_slice" not in dataset_dict:
        dataset_dict["default_group_slice"] = None

    if "app_config" not in dataset_dict:
        dataset_dict["app_config"] = {
            "media_fields": ["filepath"],
            "grid_media_field": "filepath",
            "modal_media_field": "filepath",
            "sidebar_groups": None,
            "plugins": {},
        }

    if "app_sidebar_groups" in dataset_dict:
        dataset_dict["app_config"]["sidebar_groups"] = dataset_dict.pop(
            "app_sidebar_groups"
        )

    db.datasets.replace_one(match_d, dataset_dict)


def down(db, dataset_name):
    match_d = {"name": dataset_name}
    dataset_dict = db.datasets.find_one(match_d)

    group_field = dataset_dict.pop("group_field", None)
    group_media_types = dataset_dict.pop("group_media_types", None)
    default_group_slice = dataset_dict.pop("default_group_slice", None)

    if group_field or group_media_types or default_group_slice:
        raise ValueError(
            "Cannot migrate dataset '%s' below v0.17.0 because groups were "
            "not supported before this release" % dataset_name
        )

    app_config = dataset_dict.pop("app_config", None)

    if app_config is not None and "sidebar_groups" in app_config:
        dataset_dict["app_sidebar_groups"] = app_config["sidebar_groups"]

    db.datasets.replace_one(match_d, dataset_dict)



"""
Dataset views.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import defaultdict, OrderedDict
import contextlib
from copy import copy, deepcopy
import itertools
import numbers

from bson import ObjectId
from pymongo.errors import CursorNotFound

import eta.core.utils as etau

import fiftyone.core.aggregations as foa
import fiftyone.core.collections as foc
import fiftyone.core.expressions as foe
import fiftyone.core.fields as fof
import fiftyone.core.media as fom
import fiftyone.core.odm as foo
import fiftyone.core.sample as fos
import fiftyone.core.utils as fou

fost = fou.lazy_import("fiftyone.core.stages")


class DatasetView(foc.SampleCollection):
    """A view into a :class:`fiftyone.core.dataset.Dataset`.

    Dataset views represent ordered collections of subsets of samples in a
    dataset.

    Operations on dataset views are designed to be chained together to yield
    the desired subset of the dataset, which is then iterated over to directly
    access the sample views. Each stage in the pipeline defining a dataset view
    is represented by a :class:`fiftyone.core.stages.ViewStage` instance.

    The stages of a dataset view specify:

    -   The subset of samples (and their order) that should be included
    -   The possibly-filtered fields of each sample that should be included

    Samples retrieved from dataset views are returned as
    :class:`fiftyone.core.sample.SampleView` objects, as opposed to
    :class:`fiftyone.core.sample.Sample` objects, since they may contain a
    subset of the sample's content.

    See :ref:`this page <using-views>` for an overview of working with dataset
    views.

    Args:
        dataset: the underlying :class:`fiftyone.core.dataset.Dataset` for the
            view
    """

    def __init__(
        self, dataset, _stages=None, _media_type=None, _group_slice=None
    ):
        if _stages is None:
            _stages = []

        self.__dataset = dataset
        self.__stages = _stages
        self.__media_type = _media_type
        self.__group_slice = _group_slice

    def __eq__(self, other):
        if type(other) != type(self):
            return False

        if self._root_dataset != other._root_dataset:
            return False

        # Two views into the same dataset are equal if their stage definitions
        # are equal, excluding their UUIDs
        d = self._serialize(include_uuids=False)
        other_d = other._serialize(include_uuids=False)
        return d == other_d

    def __len__(self):
        return self.count()

    def __getitem__(self, id_filepath_slice):
        if isinstance(id_filepath_slice, numbers.Integral):
            raise KeyError(
                "Accessing samples by numeric index is not supported. "
                "Use sample IDs, filepaths, slices, boolean arrays, or a "
                "boolean ViewExpression instead"
            )

        if isinstance(id_filepath_slice, slice):
            return self._slice(id_filepath_slice)

        if isinstance(id_filepath_slice, foe.ViewExpression):
            return self.match(id_filepath_slice)

        if etau.is_container(id_filepath_slice):
            return self.select(id_filepath_slice)

        try:
            oid = ObjectId(id_filepath_slice)
            query = {"_id": oid}
        except:
            oid = None
            query = {"filepath": id_filepath_slice}

        view = self.match(query)

        try:
            return next(iter(view))
        except StopIteration:
            field = "ID" if oid is not None else "filepath"
            raise KeyError(
                "No sample found with %s '%s'" % (field, id_filepath_slice)
            )

    def __copy__(self):
        return self.__class__(
            self.__dataset,
            _stages=deepcopy(self.__stages),
            _media_type=self.__media_type,
            _group_slice=self.__group_slice,
        )

    @property
    def _base_view(self):
        return self.__class__(self.__dataset)

    @property
    def _dataset(self):
        return self.__dataset

    @property
    def _root_dataset(self):
        return self.__dataset

    @property
    def _is_generated(self):
        return self._dataset._is_generated

    @property
    def _is_patches(self):
        return self._dataset._is_patches

    @property
    def _is_frames(self):
        return self._dataset._is_frames

    @property
    def _is_clips(self):
        return self._dataset._is_clips

    @property
    def _sample_cls(self):
        return fos.SampleView

    @property
    def _stages(self):
        return self.__stages

    @property
    def _all_stages(self):
        return self.__stages

    @property
    def media_type(self):
        """The media type of the view."""
        if self.__media_type is not None:
            return self.__media_type

        return self._dataset.media_type

    @property
    def group_field(self):
        """The group field of the view, or None if the view is not grouped."""
        if self.media_type != fom.GROUP:
            return None

        return self._dataset.group_field

    @property
    def group_slice(self):
        """The current group slice of the view, or None if the view is not
        grouped.
        """
        if self.media_type != fom.GROUP:
            return None

        if self.__group_slice is not None:
            return self.__group_slice

        return self._dataset.group_slice

    @group_slice.setter
    def group_slice(self, slice_name):
        if self.media_type != fom.GROUP:
            raise ValueError("DatasetView has no groups")

        if slice_name is not None and slice_name not in self.group_media_types:
            raise ValueError(
                "DatasetView has no group slice '%s'" % slice_name
            )

        self.__group_slice = slice_name

    @property
    def group_slices(self):
        """The list of group slices of the view, or None if the view is not
        grouped.
        """
        if self.media_type != fom.GROUP:
            return None

        return self._dataset.group_slices

    @property
    def group_media_types(self):
        """A dict mapping group slices to media types, or None if the view is
        not grouped.
        """
        if self.media_type != fom.GROUP:
            return None

        return self._dataset.group_media_types

    @property
    def default_group_slice(self):
        """The default group slice of the view, or None if the view is not
        grouped.
        """
        if self.media_type != fom.GROUP:
            return None

        return self._dataset.default_group_slice

    @property
    def name(self):
        """The name of the view."""
        return self.dataset_name + "-view"

    @property
    def dataset_name(self):
        """The name of the underlying dataset."""
        return self._root_dataset.name

    @property
    def info(self):
        return self._root_dataset.info

    @info.setter
    def info(self, info):
        self._root_dataset.info = info

    @property
    def app_config(self):
        return self._root_dataset.app_config

    @app_config.setter
    def app_config(self, config):
        self._root_dataset.app_config = config

    @property
    def classes(self):
        return self._root_dataset.classes

    @classes.setter
    def classes(self, classes):
        self._root_dataset.classes = classes

    @property
    def default_classes(self):
        return self._root_dataset.default_classes

    @default_classes.setter
    def default_classes(self, classes):
        self._root_dataset.default_classes = classes

    @property
    def mask_targets(self):
        return self._root_dataset.mask_targets

    @mask_targets.setter
    def mask_targets(self, targets):
        self._root_dataset.mask_targets = targets

    @property
    def default_mask_targets(self):
        return self._root_dataset.default_mask_targets

    @default_mask_targets.setter
    def default_mask_targets(self, targets):
        self._root_dataset.default_mask_targets = targets

    @property
    def skeletons(self):
        return self._root_dataset.skeletons

    @skeletons.setter
    def skeletons(self, skeletons):
        self._root_dataset.skeletons = skeletons

    @property
    def default_skeleton(self):
        return self._root_dataset.default_skeleton

    @default_skeleton.setter
    def default_skeleton(self, skeleton):
        self._root_dataset.default_skeleton = skeleton

    def summary(self):
        """Returns a string summary of the view.

        Returns:
            a string summary
        """
        elements = [
            ("Dataset:", self.dataset_name),
            ("Media type:", self.media_type),
            ("Num %s:" % self._elements_str, self.count()),
        ]

        if self.media_type == fom.GROUP:
            elements.insert(2, ("Group slice:", self.group_slice))

        elements = fou.justify_headings(elements)
        lines = ["%s %s" % tuple(e) for e in elements]

        lines.extend(
            [
                "%s fields:" % self._element_str.capitalize(),
                self._to_fields_str(self.get_field_schema()),
            ]
        )

        if self._has_frame_fields():
            lines.extend(
                [
                    "Frame fields:",
                    self._to_fields_str(self.get_frame_field_schema()),
                ]
            )

        lines.extend(["View stages:", self._make_view_stages_str()])

        return "\n".join(lines)

    def _make_view_stages_str(self):
        if not self._all_stages:
            return "    ---"

        return "    " + "\n    ".join(
            [
                "%d. %s" % (idx, str(stage))
                for idx, stage in enumerate(self._all_stages, 1)
            ]
        )

    def view(self):
        """Returns a copy of this view.

        Returns:
            a :class:`DatasetView`
        """
        return copy(self)

    def iter_samples(self, progress=False, autosave=False, batch_size=None):
        """Returns an iterator over the samples in the view.

        Examples::

            import random as r
            import string as s

            import fiftyone as fo
            import fiftyone.zoo as foz

            dataset = foz.load_zoo_dataset("cifar10", split="test")
            view = dataset.shuffle().limit(5000)

            def make_label():
                return "".join(r.choice(s.ascii_letters) for i in range(10))

            # No save context
            for sample in view.iter_samples(progress=True):
                sample.ground_truth.label = make_label()
                sample.save()

            # Save in batches of 10
            for sample in view.iter_samples(
                progress=True, autosave=True, batch_size=10
            ):
                sample.ground_truth.label = make_label()

            # Save every 0.5 seconds
            for sample in view.iter_samples(
                progress=True, autosave=True, batch_size=0.5
            ):
                sample.ground_truth.label = make_label()

        Args:
            progress (False): whether to render a progress bar tracking the
                iterator's progress
            autosave (False): whether to automatically save changes to samples
                emitted by this iterator
            batch_size (None): a batch size to use when autosaving samples. Can
                either be an integer specifying the number of samples to save
                in a batch, or a float number of seconds between batched saves

        Returns:
            an iterator over :class:`fiftyone.core.sample.SampleView` instances
        """
        with contextlib.ExitStack() as exit_context:
            samples = self._iter_samples()

            if progress:
                pb = fou.ProgressBar(total=len(self))
                exit_context.enter_context(pb)
                samples = pb(samples)

            if autosave:
                save_context = foc.SaveContext(self, batch_size=batch_size)
                exit_context.enter_context(save_context)

            for sample in samples:
                yield sample

                if autosave:
                    save_context.save(sample)

    def _iter_samples(self):
        make_sample = self._make_sample_fcn()
        index = 0

        try:
            for d in self._aggregate(detach_frames=True, detach_groups=True):
                sample = make_sample(d)

                index += 1
                yield sample
        except CursorNotFound:
            # The cursor has timed out so we yield from a new one after
            # skipping to the last offset
            view = self.skip(index)
            for sample in view._iter_samples():
                yield sample

    def _make_sample_fcn(self):
        sample_cls = self._sample_cls
        selected_fields, excluded_fields = self._get_selected_excluded_fields(
            roots_only=True
        )
        filtered_fields = self._get_filtered_fields()

        def make_sample(d):
            try:
                doc = self._dataset._sample_dict_to_doc(d)
                return sample_cls(
                    doc,
                    self,
                    selected_fields=selected_fields,
                    excluded_fields=excluded_fields,
                    filtered_fields=filtered_fields,
                )
            except Exception as e:
                raise ValueError(
                    "Failed to load sample from the database. This is likely "
                    "due to an invalid stage in the DatasetView"
                ) from e

        return make_sample

    def iter_groups(self, progress=False, autosave=False, batch_size=None):
        """Returns an iterator over the groups in the view.

        Examples::

            import random as r
            import string as s

            import fiftyone as fo
            import fiftyone.zoo as foz

            dataset = foz.load_zoo_dataset("quickstart-groups")
            view = dataset.select_fields()

            def make_label():
                return "".join(r.choice(s.ascii_letters) for i in range(10))

            # No save context
            for group in view.iter_groups(progress=True):
                for sample in group.values():
                    sample["test"] = make_label()
                    sample.save()

            # Save in batches of 10
            for group in view.iter_groups(
                progress=True, autosave=True, batch_size=10
            ):
                for sample in group.values():
                    sample["test"] = make_label()

            # Save every 0.5 seconds
            for group in view.iter_groups(
                progress=True, autosave=True, batch_size=0.5
            ):
                for sample in group.values():
                    sample["test"] = make_label()

        Args:
            progress (False): whether to render a progress bar tracking the
                iterator's progress
            autosave (False): whether to automatically save changes to samples
                emitted by this iterator
            batch_size (None): a batch size to use when autosaving samples. Can
                either be an integer specifying the number of samples to save
                in a batch, or a float number of seconds between batched saves

        Returns:
            an iterator that emits dicts mapping slice names to
            :class:`fiftyone.core.sample.SampleView` instances, one per group
        """
        if self.media_type != fom.GROUP:
            raise ValueError("%s does not contain groups" % type(self))

        with contextlib.ExitStack() as exit_context:
            groups = self._iter_groups()

            if progress:
                pb = fou.ProgressBar(total=len(self))
                exit_context.enter_context(pb)
                groups = pb(groups)

            if autosave:
                save_context = foc.SaveContext(self, batch_size=batch_size)
                exit_context.enter_context(save_context)

            for group in groups:
                yield group

                if autosave:
                    for sample in group.values():
                        save_context.save(sample)

    def _iter_groups(self):
        make_sample = self._make_sample_fcn()
        index = 0

        group_field = self.group_field
        curr_id = None
        group = {}

        try:
            for d in self._aggregate(groups_only=True, detach_frames=True):
                sample = make_sample(d)

                group_id = sample[group_field].id
                if curr_id is None:
                    # First overall element
                    curr_id = group_id
                    group[sample[group_field].name] = sample
                elif group_id == curr_id:
                    # Add element to group
                    group[sample[group_field].name] = sample
                else:
                    # Flush last group
                    index += 1
                    yield group

                    # First element of new group
                    curr_id = group_id
                    group = {}
                    group[sample[group_field].name] = sample

            if group:
                yield group
        except CursorNotFound:
            # The cursor has timed out so we yield from a new one after
            # skipping to the last offset
            view = self.skip(index)
            for group in view._iter_groups():
                yield group

    def get_group(self, group_id):
        """Returns a dict containing the samples for the given group ID.

        Args:
            group_id: a group ID

        Returns:
            a dict mapping group names to
            :class:`fiftyone.core.sample.SampleView` instances

        Raises:
            KeyError: if the group ID is not found
        """
        if self.media_type != fom.GROUP:
            raise ValueError("%s does not contain groups" % type(self))

        if self.group_field is None:
            raise ValueError("%s has no group field" % type(self))

        group_field = self.group_field
        id_field = group_field + "._id"

        view = self.match(foe.ViewField(id_field) == ObjectId(group_id))

        try:
            return next(iter(view._iter_groups()))
        except StopIteration:
            raise KeyError(
                "No group found with ID '%s' in field '%s'"
                % (group_id, group_field)
            )

    def get_field_schema(
        self,
        ftype=None,
        embedded_doc_type=None,
        include_private=False,
        flat=False,
    ):
        """Returns a schema dictionary describing the fields of the samples in
        the view.

        Args:
            ftype (None): an optional field type to which to restrict the
                returned schema. Must be a subclass of
                :class:`fiftyone.core.fields.Field`
            embedded_doc_type (None): an optional embedded document type to
                which to restrict the returned schema. Must be a subclass of
                :class:`fiftyone.core.odm.BaseEmbeddedDocument`
            include_private (False): whether to include fields that start with
                ``_`` in the returned schema
            flat (False): whether to return a flattened schema where all
                embedded document fields are included as top-level keys

        Returns:
             a dictionary mapping field names to field types
        """
        schema = self._dataset.get_field_schema(
            ftype=ftype,
            embedded_doc_type=embedded_doc_type,
            include_private=include_private,
        )

        schema = self._get_filtered_schema(schema)

        if flat:
            schema = fof.flatten_schema(
                schema,
                ftype=ftype,
                embedded_doc_type=embedded_doc_type,
                include_private=include_private,
            )

        return schema

    def get_frame_field_schema(
        self,
        ftype=None,
        embedded_doc_type=None,
        include_private=False,
        flat=False,
    ):
        """Returns a schema dictionary describing the fields of the frames of
        the samples in the view.

        Only applicable for views that contain videos.

        Args:
            ftype (None): an optional field type to which to restrict the
                returned schema. Must be a subclass of
                :class:`fiftyone.core.fields.Field`
            embedded_doc_type (None): an optional embedded document type to
                which to restrict the returned schema. Must be a subclass of
                :class:`fiftyone.core.odm.BaseEmbeddedDocument`
            include_private (False): whether to include fields that start with
                ``_`` in the returned schema
            flat (False): whether to return a flattened schema where all
                embedded document fields are included as top-level keys

        Returns:
            a dictionary mapping field names to field types, or ``None`` if
            the view does not contain videos
        """
        if not self._has_frame_fields():
            return None

        schema = self._dataset.get_frame_field_schema(
            ftype=ftype,
            embedded_doc_type=embedded_doc_type,
            include_private=include_private,
        )

        schema = self._get_filtered_schema(schema, frames=True)

        if flat:
            schema = fof.flatten_schema(
                schema,
                ftype=ftype,
                embedded_doc_type=embedded_doc_type,
                include_private=include_private,
            )

        return schema

    def clone_sample_field(self, field_name, new_field_name):
        """Clones the given sample field of the view into a new field of the
        dataset.

        You can use dot notation (``embedded.field.name``) to clone embedded
        fields.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.

        .. warning::

            If ``new_field_name`` is an embedded field, be aware that this
            operation will save the entire top-level field of
            ``new_field_name`` after performing the clone, which may result in
            data modification/loss if this view modifies this field in any
            other ways.

        Args:
            field_name: the field name or ``embedded.field.name``
            new_field_name: the new field name or ``embedded.field.name``
        """
        self._dataset._clone_sample_fields(
            {field_name: new_field_name}, view=self
        )

    def clone_sample_fields(self, field_mapping):
        """Clones the given sample fields of the view into new fields of the
        dataset.

        You can use dot notation (``embedded.field.name``) to clone embedded
        fields.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.

        .. warning::

            If any of the new field names to specify are embedded fields, be
            aware that this operation will save the entire top-level new
            fields after performing the clone, which may result in data
            modification/loss if this view modifies these fields in any other
            ways.

        Args:
            field_mapping: a dict mapping field names to new field names into
                which to clone each field
        """
        self._dataset._clone_sample_fields(field_mapping, view=self)

    def clone_frame_field(self, field_name, new_field_name):
        """Clones the frame-level field of the view into a new field.

        You can use dot notation (``embedded.field.name``) to clone embedded
        frame fields.

        Only applicable to views that contain videos.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.

        .. warning::

            If ``new_field_name`` is an embedded field, be aware that this
            operation will save the entire top-level field of
            ``new_field_name`` after performing the clone, which may result in
            data modification/loss if this view modifies this field in any
            other ways.

        Args:
            field_name: the field name or ``embedded.field.name``
            new_field_name: the new field name or ``embedded.field.name``
        """
        self._dataset._clone_frame_fields(
            {field_name: new_field_name}, view=self
        )

    def clone_frame_fields(self, field_mapping):
        """Clones the frame-level fields of the view into new frame-level
        fields of the dataset.

        You can use dot notation (``embedded.field.name``) to clone embedded
        frame fields.

        Only applicable to views that contain videos.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.

        .. warning::

            If any of the new field names to specify are embedded fields, be
            aware that this operation will save the entire top-level new
            fields after performing the clone, which may result in data
            modification/loss if this view modifies these fields in any other
            ways.

        Args:
            field_mapping: a dict mapping field names to new field names into
                which to clone each field
        """
        self._dataset._clone_frame_fields(field_mapping, view=self)

    def clear_sample_field(self, field_name):
        """Clears the values of the field from all samples in the view.

        The field will remain in the dataset's schema, and all samples in the
        view will have the value ``None`` for the field.

        You can use dot notation (``embedded.field.name``) to clear embedded
        fields.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.

        .. warning::

            If the field name you specify is an embedded field, be aware that
            this operation will save the entire top-level field after clearing
            the field, which may result in data modification/loss if this view
            modifies the field in any other ways.

        Args:
            field_name: the field name or ``embedded.field.name``
        """
        self._dataset._clear_sample_fields(field_name, view=self)

    def clear_sample_fields(self, field_names):
        """Clears the values of the fields from all samples in the view.

        The fields will remain in the dataset's schema, and all samples in the
        view will have the value ``None`` for the fields.

        You can use dot notation (``embedded.field.name``) to clear embedded
        fields.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.

        .. warning::

            If any of the field names you specify are embedded fields, be aware
            that this operation will save the entire top-level field after
            clearing the fields, which may result in data modification/loss if
            this view modifies these fields in any other ways.

        Args:
            field_names: the field name or iterable of field names
        """
        self._dataset._clear_sample_fields(field_names, view=self)

    def clear_frame_field(self, field_name):
        """Clears the values of the frame-level field from all samples in the
        view.

        The field will remain in the dataset's frame schema, and all frames in
        the view will have the value ``None`` for the field.

        You can use dot notation (``embedded.field.name``) to clear embedded
        frame fields.

        Only applicable to views that contain videos.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.

        .. warning::

            If the field name you specify is an embedded field, be aware that
            this operation will save the entire top-level field after clearing
            the field, which may result in data modification/loss if this view
            modifies the field in any other ways.

        Args:
            field_name: the field name or ``embedded.field.name``
        """
        self._dataset._clear_frame_fields(field_name, view=self)

    def clear_frame_fields(self, field_names):
        """Clears the values of the frame-level fields from all samples in the
        view.

        The fields will remain in the dataset's frame schema, and all frames in
        the view will have the value ``None`` for the fields.

        You can use dot notation (``embedded.field.name``) to clear embedded
        frame fields.

        Only applicable to views that contain videos.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.

        .. warning::

            If any of the field names you specify are embedded fields, be aware
            that this operation will save the entire top-level field after
            clearing the fields, which may result in data modification/loss if
            this view modifies these fields in any other ways.

        Args:
            field_names: the field name or iterable of field names
        """
        self._dataset._clear_frame_fields(field_names, view=self)

    def clear(self):
        """Deletes all samples in the view from the underlying dataset.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.
        """
        self._dataset._clear(view=self)

    def clear_frames(self):
        """Deletes all frame labels from the samples in the view from the
        underlying dataset.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.
        """
        self._dataset._clear_frames(view=self)

    def keep(self):
        """Deletes all samples that are **not** in the view from the underlying
        dataset.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.
        """
        self._dataset._keep(view=self)

    def keep_fields(self):
        """Deletes all fields that are excluded from the view from the
        underlying dataset.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.
        """
        self._dataset._keep_fields(view=self)

    def keep_frames(self):
        """For each sample in the view, deletes all frames labels that are
        **not** in the view from the underlying dataset.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.
        """
        self._dataset._keep_frames(view=self)

    def ensure_frames(self):
        """Ensures that the video view contains frame instances for every frame
        of each sample's source video.

        Empty frames will be inserted for missing frames, and already existing
        frames are left unchanged.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.
        """
        self._dataset._ensure_frames(view=self)

    def save(self, fields=None):
        """Saves the contents of the view to the database.

        This method **does not** delete samples or frames from the underlying
        dataset that this view excludes.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.

        .. warning::

            If a view has excluded fields or filtered list values, this method
            will permanently delete this data from the dataset, unless
            ``fields`` is used to omit such fields from the save.

        Args:
            fields (None): an optional field or list of fields to save. If
                specified, only these field's contents are modified
        """
        self._dataset._save(view=self, fields=fields)

    def clone(self, name=None, persistent=False):
        """Creates a new dataset containing a copy of the contents of the view.

        Dataset clones contain deep copies of all samples and dataset-level
        information in the source collection. The source *media files*,
        however, are not copied.

        Args:
            name (None): a name for the cloned dataset. By default,
                :func:`get_default_dataset_name` is used
            persistent (False): whether the cloned dataset should be persistent

        Returns:
            the new :class:`fiftyone.core.dataset.Dataset`
        """
        return self._dataset._clone(
            name=name,
            persistent=persistent,
            view=self,
        )

    def reload(self):
        """Reloads the underlying dataset from the database.

        Note that :class:`fiftyone.core.sample.SampleView` instances are not
        singletons, so any in-memory samples extracted from this view will not
        be updated by calling this method.
        """
        self._dataset.reload()

    def to_dict(
        self,
        rel_dir=None,
        include_private=False,
        include_frames=False,
        frame_labels_dir=None,
        pretty_print=False,
    ):
        """Returns a JSON dictionary representation of the view.

        Args:
            rel_dir (None): a relative directory to remove from the
                ``filepath`` of each sample, if possible. The path is converted
                to an absolute path (if necessary) via
                :func:`fiftyone.core.utils.normalize_path`. The typical use
                case for this argument is that your source data lives in a
                single directory and you wish to serialize relative, rather
                than absolute, paths to the data within that directory
            include_private (False): whether to include private fields
            include_frames (False): whether to include the frame labels for
                video samples
            frame_labels_dir (None): a directory in which to write per-sample
                JSON files containing the frame labels for video samples. If
                omitted, frame labels will be included directly in the returned
                JSON dict (which can be quite quite large for video datasets
                containing many frames). Only applicable to datasets that
                contain videos when ``include_frames`` is True
            pretty_print (False): whether to render frame labels JSON in human
                readable format with newlines and indentations. Only applicable
                to datasets that contain videos when a ``frame_labels_dir`` is
                provided

        Returns:
            a JSON dict
        """
        d = super().to_dict(
            rel_dir=rel_dir,
            include_private=include_private,
            include_frames=include_frames,
            frame_labels_dir=frame_labels_dir,
            pretty_print=pretty_print,
        )
        samples = d.pop("samples")  # hack so that `samples` is last in JSON
        d["stages"] = self._serialize(include_uuids=False)
        d["samples"] = samples
        return d

    def _needs_frames(self):
        dataset = self._dataset

        if not dataset._has_frame_fields():
            return False

        for stage in self._stages:
            if stage._needs_frames(dataset):
                return True

        return False

    def _pipeline(
        self,
        pipeline=None,
        media_type=None,
        attach_frames=False,
        detach_frames=False,
        frames_only=False,
        support=None,
        group_slice=None,
        group_slices=None,
        groups_only=False,
        detach_groups=False,
        manual_group_select=False,
        post_pipeline=None,
    ):
        _pipelines = []
        _view = self._base_view

        _contains_videos = self._dataset._contains_videos(any_slice=True)
        _found_select_group_slice = False
        _attach_frames_idx = None
        _attach_frames_idx0 = None

        _contains_groups = self._dataset.media_type == fom.GROUP
        _group_slices = set()
        _attach_groups_idx = None

        for idx, stage in enumerate(self._stages):
            if isinstance(stage, fost.SelectGroupSlices):
                # We might need to reattach frames after `SelectGroupSlices`,
                # since it involves a `$lookup` that resets the samples
                _found_select_group_slice = True
                _attach_frames_idx0 = _attach_frames_idx
                _attach_frames_idx = None

            # Determine if stage needs frames attached
            if (
                _contains_videos
                and _attach_frames_idx is None
                and stage._needs_frames(_view)
            ):
                _attach_frames_idx = idx

            if _contains_groups:
                # Special case: report a manual override if the first stage
                # transforms a grouped collection into a non-grouped collection
                if idx == 0:
                    _media_type = stage.get_media_type(_view)
                    if _media_type not in (None, fom.GROUP):
                        manual_group_select = True

                # Determine if stage needs group slices attached
                _stage_group_slices = stage._needs_group_slices(_view)
                if _stage_group_slices:
                    if _attach_groups_idx is None:
                        _attach_groups_idx = idx

                    _group_slices.update(_stage_group_slices)

            # Generate stage's pipeline
            _pipelines.append(stage.to_mongo(_view))
            _view = _view.add_stage(stage)

        if _attach_frames_idx is None and (attach_frames or frames_only):
            _attach_frames_idx = len(_pipelines)

        #######################################################################
        # Insert frame lookup pipeline(s) if needed
        #######################################################################

        if _attach_frames_idx0 is not None and _attach_frames_idx is not None:
            # Two lookups are required; manually do the **last** one and rely
            # on dataset._pipeline() to do the first one
            attach_frames = True
            _pipeline = self._dataset._attach_frames_pipeline(support=support)
            _pipelines.insert(_attach_frames_idx, _pipeline)
        elif _found_select_group_slice and _attach_frames_idx is not None:
            # Must manually attach frames after the group selection
            attach_frames = None  # special syntax: frames already attached
            _pipeline = self._dataset._attach_frames_pipeline(support=support)
            _pipelines.insert(_attach_frames_idx, _pipeline)
        elif _attach_frames_idx0 is not None or _attach_frames_idx is not None:
            # Exactly one lookup is required; rely on dataset._pipeline() to
            # do it
            attach_frames = True

        # @todo use the optimization below instead, which injects frames as
        # late as possible in the pipeline. We can't currently use it because
        # there's some issue with poster frames in the App if the frames are
        # not attached first...

        """
        if _attach_frames_idx0 is not None or _attach_frames_idx is not None:
            attach_frames = None  # special syntax: frames already attached

            if _attach_frames_idx0 is not None:
                _pipeline = self._dataset._attach_frames_pipeline(
                    support=support
                )
                _pipelines.insert(_attach_frames_idx0, _pipeline)

            if _attach_frames_idx is not None:
                if _attach_frames_idx0 is not None:
                    _attach_frames_idx += 1

                _pipeline = self._dataset._attach_frames_pipeline(
                    support=support
                )
                _pipelines.insert(_attach_frames_idx, _pipeline)
        """

        #######################################################################

        # Insert group lookup pipline if needed
        if _attach_groups_idx is not None:
            if group_slices:
                _group_slices.update(group_slices)

            group_slices = None
            _pipeline = self._dataset._attach_groups_pipeline(
                group_slices=_group_slices
            )
            _pipelines.insert(_attach_groups_idx, _pipeline)

        if pipeline is not None:
            _pipelines.append(pipeline)

        _pipeline = list(itertools.chain.from_iterable(_pipelines))

        if media_type is None:
            media_type = self.media_type

        if group_slice is None:
            group_slice = self.group_slice

        return self._dataset._pipeline(
            pipeline=_pipeline,
            attach_frames=attach_frames,
            detach_frames=detach_frames,
            frames_only=frames_only,
            support=support,
            media_type=media_type,
            group_slice=group_slice,
            group_slices=group_slices,
            groups_only=groups_only,
            detach_groups=detach_groups,
            manual_group_select=manual_group_select,
            post_pipeline=post_pipeline,
        )

    def _aggregate(
        self,
        pipeline=None,
        media_type=None,
        attach_frames=False,
        detach_frames=False,
        frames_only=False,
        support=None,
        group_slice=None,
        group_slices=None,
        groups_only=False,
        detach_groups=False,
        manual_group_select=False,
        post_pipeline=None,
    ):
        _pipeline = self._pipeline(
            pipeline=pipeline,
            media_type=media_type,
            attach_frames=attach_frames,
            detach_frames=detach_frames,
            frames_only=frames_only,
            support=support,
            group_slice=group_slice,
            group_slices=group_slices,
            groups_only=groups_only,
            detach_groups=detach_groups,
            manual_group_select=manual_group_select,
            post_pipeline=post_pipeline,
        )

        return foo.aggregate(self._dataset._sample_collection, _pipeline)

    def _serialize(self, include_uuids=True):
        return [
            stage._serialize(include_uuid=include_uuids)
            for stage in self._all_stages
        ]

    @staticmethod
    def _build(dataset, stage_dicts):
        view = dataset.view()
        for stage_dict in stage_dicts:
            stage = fost.ViewStage._from_dict(stage_dict)
            view = view.add_stage(stage)

        return view

    def _slice(self, s):
        if s.step is not None and s.step != 1:
            raise ValueError(
                "Unsupported slice '%s'; step is not supported" % s
            )

        _len = None

        start = s.start
        if start is not None:
            if start < 0:
                _len = len(self)
                start += _len

            if start <= 0:
                start = None

        stop = s.stop
        if stop is not None and stop < 0:
            if _len is None:
                _len = len(self)

            stop += _len

        if start is None:
            if stop is None:
                return self

            return self.limit(stop)

        if stop is None:
            return self.skip(start)

        return self.skip(start).limit(stop - start)

    def _add_view_stage(self, stage):
        stage.validate(self)

        if stage.has_view:
            view = stage.load_view(self)
        else:
            view = copy(self)
            view._stages.append(stage)

            media_type = stage.get_media_type(self)
            if media_type is not None:
                view._set_media_type(media_type)

        return view

    def _set_media_type(self, media_type):
        self.__media_type = media_type

        if media_type != fom.GROUP:
            self.__group_slice = None

    def _get_filtered_schema(self, schema, frames=False):
        if schema is None:
            return None

        selected_fields, excluded_fields = self._get_selected_excluded_fields(
            frames=frames
        )

        if selected_fields is not None or excluded_fields is not None:
            _filter_schema(schema, selected_fields, excluded_fields)

        return schema

    def _get_selected_excluded_fields(self, frames=False, roots_only=False):
        selected_fields = None
        excluded_fields = None

        dataset = self._dataset
        for stage in self._stages:
            sf = stage.get_selected_fields(dataset, frames=frames)
            if sf:
                if roots_only:
                    sf = {f.split(".", 1)[0] for f in sf}

                if selected_fields is None:
                    selected_fields = set(sf)
                else:
                    selected_fields.intersection_update(sf)

            ef = stage.get_excluded_fields(dataset, frames=frames)
            if ef:
                if roots_only:
                    ef = {f for f in ef if "." not in f}

                if excluded_fields is None:
                    excluded_fields = set(ef)
                else:
                    excluded_fields.update(ef)

        if (
            roots_only
            and selected_fields is not None
            and excluded_fields is not None
        ):
            selected_fields.difference_update(excluded_fields)
            excluded_fields = None

        return selected_fields, excluded_fields

    def _get_filtered_fields(self, frames=False):
        filtered_fields = None

        dataset = self._dataset
        for stage in self._stages:
            ff = stage.get_filtered_fields(dataset, frames=frames)
            if ff:
                if filtered_fields is None:
                    filtered_fields = set(ff)
                else:
                    filtered_fields.update(ff)

        return filtered_fields

    def _get_missing_fields(self, frames=False):
        if frames:
            if not self._has_frame_fields():
                return set()

            dataset_schema = self._dataset.get_frame_field_schema()
            view_schema = self.get_frame_field_schema()
        else:
            dataset_schema = self._dataset.get_field_schema()
            view_schema = self.get_field_schema()

        return set(dataset_schema.keys()) - set(view_schema.keys())

    def _contains_all_fields(self, frames=False):
        selected_fields, excluded_fields = self._get_selected_excluded_fields(
            frames=frames
        )
        filtered_fields = self._get_filtered_fields(frames=frames)
        return not any((selected_fields, excluded_fields, filtered_fields))

    def _get_group_media_types(self):
        for stage in reversed(self._stages):
            if isinstance(stage, fost.SelectGroupSlices):
                return stage._get_group_media_types(self._dataset)

        return self._dataset.group_media_types


def make_optimized_select_view(
    sample_collection,
    sample_ids,
    ordered=False,
    groups=False,
    select_groups=False,
):
    """Returns a view that selects the provided sample IDs that is optimized
    to reduce the document list as early as possible in the pipeline.

    .. warning::

        This method **deletes** any other view stages that reorder/select
        documents, so the returned view may not respect the order of the
        documents in the input collection.

    Args:
        sample_collection:  a
            :class:`fiftyone.core.collections.SampleCollection`
        sample_ids: a sample ID or iterable of sample IDs to select
        ordered (False): whether to sort the samples in the returned view to
            match the order of the provided IDs
        groups (False): whether the IDs are group IDs, not sample IDs
        select_groups (False): whether to select sample groups via sample ids

    Returns:
        a :class:`DatasetView`
    """
    view = sample_collection.view()

    if any(isinstance(stage, fost.Mongo) for stage in view._stages):
        #
        # We have no way of knowing what a `Mongo()` stage might do, so we must
        # run the entire view's aggregation first and then select the samples
        # of interest at the end
        #
        if groups:
            return view.select_groups(sample_ids, ordered=ordered)
        elif view.media_type == fom.GROUP and not select_groups:
            return view.select_group_slices(_allow_mixed=True)

        view = view.select(sample_ids, ordered=ordered)

        if view.media_type == fom.GROUP and select_groups:
            view = view.select_group_slices(_allow_mixed=True)

        return view

    #
    # Selecting the samples of interest first can be significantly faster than
    # running the entire aggregation and then selecting them.
    #
    # However, in order to do that, we must omit any `Skip()` stages, which
    # depend on the number of documents in the pipeline.
    #
    # In addition, we take the liberty of omitting other stages that are known
    # to only select/reorder documents.
    #
    # @note this is brittle because if any new stages like `Skip()` are added
    # that could affect our ability to select the samples of interest first,
    # we'll need to account for that here...
    #

    optimized_view = view._base_view

    if groups:
        optimized_view = optimized_view.select_groups(
            sample_ids, ordered=ordered
        )
    else:
        if view.media_type == fom.GROUP and not select_groups:
            optimized_view = optimized_view.select_group_slices(
                _allow_mixed=True
            )

        optimized_view = optimized_view.select(sample_ids, ordered=ordered)
        if view.media_type == fom.GROUP and select_groups:
            optimized_view = optimized_view.select_group_slices(
                _allow_mixed=True
            )

    for stage in view._stages:
        if type(stage) not in fost._STAGES_THAT_SELECT_OR_REORDER:
            optimized_view._stages.append(stage)

    return optimized_view


def _filter_schema(schema, selected_fields, excluded_fields):
    selected_fields, roots1 = _parse_selected_fields(selected_fields)
    excluded_fields, roots2 = _parse_excluded_fields(excluded_fields)
    filtered_roots = roots1 | roots2

    # Explicitly include roots of any embedded fields that have been selected
    if roots1:
        selected_fields[""].update(roots1)

    # Copy any top-level fields whose embedded fields will be filtered
    if filtered_roots:
        for name in tuple(schema.keys()):
            if name in filtered_roots:
                schema[name] = schema[name].copy()

    sf = selected_fields.get("", None)
    ef = excluded_fields.get("", None)

    if sf is not None:
        for name in tuple(schema.keys()):
            if name not in sf:
                del schema[name]

    if ef is not None:
        for name in tuple(schema.keys()):
            if name in ef:
                del schema[name]

    if filtered_roots:
        for name, field in schema.items():
            _filter_embedded_field_schema(
                field, name, selected_fields, excluded_fields
            )


def _parse_selected_fields(paths):
    d = defaultdict(set)
    r = set()

    if paths is not None:
        for path in paths:
            if "." in path:
                chunks = path.split(".")
                root = chunks[0]
                r.add(root)

                for i in range(1, len(chunks)):
                    base = ".".join(chunks[:i])
                    leaf = chunks[i]
                    d[base].add(leaf)
            else:
                d[""].add(path)

    return d, r


def _parse_excluded_fields(paths):
    d = defaultdict(set)
    r = set()

    if paths is not None:
        for path in paths:
            if "." in path:
                root = path.split(".", 1)[0]
                r.add(root)

                base, leaf = path.rsplit(".", 1)
                d[base].add(leaf)
            else:
                d[""].add(path)

    return d, r


def _filter_embedded_field_schema(
    field, path, selected_fields, excluded_fields
):
    while isinstance(field, fof.ListField):
        field = field.field

    if not isinstance(field, fof.EmbeddedDocumentField):
        return

    sf = selected_fields.get(path, None)
    ef = excluded_fields.get(path, None)

    if sf is not None or ef is not None:
        field._use_view(selected_fields=sf, excluded_fields=ef)

    for name, _field in field._fields.items():
        _path = path + "." + name
        _filter_embedded_field_schema(
            _field, _path, selected_fields, excluded_fields
        )



"""
Video frame views.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import defaultdict
from copy import deepcopy
import logging
import os

from bson import ObjectId
from pymongo import UpdateOne

import eta.core.utils as etau

import fiftyone as fo
import fiftyone.core.dataset as fod
import fiftyone.core.fields as fof
import fiftyone.core.media as fom
import fiftyone.core.sample as fos
import fiftyone.core.odm as foo
import fiftyone.core.odm.sample as foos
import fiftyone.core.utils as fou
import fiftyone.core.validation as fova
import fiftyone.core.view as fov

fouv = fou.lazy_import("fiftyone.utils.video")


logger = logging.getLogger(__name__)


class FrameView(fos.SampleView):
    """A frame in a :class:`FramesView`.

    :class:`FrameView` instances should not be created manually; they are
    generated by iterating over :class:`FramesView` instances.

    Args:
        doc: a :class:`fiftyone.core.odm.DatasetSampleDocument`
        view: the :class:`FramesView` that the frame belongs to
        selected_fields (None): a set of field names that this view is
            restricted to
        excluded_fields (None): a set of field names that are excluded from
            this view
        filtered_fields (None): a set of field names of list fields that are
            filtered in this view
    """

    @property
    def _sample_id(self):
        return ObjectId(self._doc.sample_id)

    def _save(self, deferred=False):
        if deferred:
            raise NotImplementedError(
                "Frames views do not support save contexts"
            )

        super()._save(deferred=deferred)
        self._view._sync_source_sample(self)


class FramesView(fov.DatasetView):
    """A :class:`fiftyone.core.view.DatasetView` of frames from a video
    :class:`fiftyone.core.dataset.Dataset`.

    Frames views contain an ordered collection of frames, each of which
    corresponds to a single frame of a video from the source collection.

    Frames retrieved from frames views are returned as :class:`FrameView`
    objects.

    Args:
        source_collection: the
            :class:`fiftyone.core.collections.SampleCollection` from which this
            view was created
        frames_stage: the :class:`fiftyone.core.stages.ToFrames` stage that
            defines how the frames were created
        frames_dataset: the :class:`fiftyone.core.dataset.Dataset` that serves
            the frames in this view
    """

    def __init__(
        self, source_collection, frames_stage, frames_dataset, _stages=None
    ):
        if _stages is None:
            _stages = []

        self._source_collection = source_collection
        self._frames_stage = frames_stage
        self._frames_dataset = frames_dataset
        self.__stages = _stages

    def __copy__(self):
        return self.__class__(
            self._source_collection,
            deepcopy(self._frames_stage),
            self._frames_dataset,
            _stages=deepcopy(self.__stages),
        )

    @property
    def _base_view(self):
        return self.__class__(
            self._source_collection,
            self._frames_stage,
            self._frames_dataset,
        )

    @property
    def _dataset(self):
        return self._frames_dataset

    @property
    def _root_dataset(self):
        return self._source_collection._root_dataset

    @property
    def _sample_cls(self):
        return FrameView

    @property
    def _stages(self):
        return self.__stages

    @property
    def _all_stages(self):
        return (
            self._source_collection.view()._all_stages
            + [self._frames_stage]
            + self.__stages
        )

    @property
    def name(self):
        return self.dataset_name + "-frames"

    @property
    def media_type(self):
        return fom.IMAGE

    def _get_sample_only_fields(
        self, include_private=False, use_db_fields=False
    ):
        sample_only_fields = set(
            self._get_default_sample_fields(
                include_private=include_private, use_db_fields=use_db_fields
            )
        )

        # If sample_frames != dynamic, `filepath` can be synced
        config = self._frames_stage.config or {}
        if config.get("sample_frames", None) != "dynamic":
            sample_only_fields.discard("filepath")

        return sample_only_fields

    def _tag_labels(self, tags, label_field, ids=None, label_ids=None):
        ids, label_ids = super()._tag_labels(
            tags, label_field, ids=ids, label_ids=label_ids
        )

        frame_field = self._source_collection._FRAMES_PREFIX + label_field
        self._source_collection._tag_labels(
            tags, frame_field, ids=ids, label_ids=label_ids
        )

    def _untag_labels(self, tags, label_field, ids=None, label_ids=None):
        ids, label_ids = super()._untag_labels(
            tags, label_field, ids=ids, label_ids=label_ids
        )

        frame_field = self._source_collection._FRAMES_PREFIX + label_field
        self._source_collection._untag_labels(
            tags, frame_field, ids=ids, label_ids=label_ids
        )

    def set_values(self, field_name, *args, **kwargs):
        # The `set_values()` operation could change the contents of this view,
        # so we first record the sample IDs that need to be synced
        if self._stages:
            ids = self.values("id")
        else:
            ids = None

        super().set_values(field_name, *args, **kwargs)

        field = field_name.split(".", 1)[0]
        self._sync_source(fields=[field], ids=ids)

    def save(self, fields=None):
        """Saves the frames in this view to the underlying dataset.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.

        .. warning::

            This will permanently delete any omitted or filtered contents from
            the frames of the source dataset.

        Args:
            fields (None): an optional field or list of fields to save. If
                specified, only these fields are overwritten
        """
        if etau.is_str(fields):
            fields = [fields]

        super().save(fields=fields)

        self._sync_source(fields=fields)

    def keep(self):
        """Deletes all frames that are **not** in this view from the underlying
        dataset.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.
        """

        # The `keep()` operation below will delete frames, so we must sync
        # deletions to the source dataset first
        self._sync_source(update=False, delete=True)

        super().keep()

    def keep_fields(self):
        """Deletes any sample fields that have been excluded in this view from
        the frames of the underlying dataset.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.
        """
        self._sync_source_keep_fields()

        super().keep_fields()

    def reload(self):
        """Reloads this view from the source collection in the database.

        Note that :class:`FrameView` instances are not singletons, so any
        in-memory frames extracted from this view will not be updated by
        calling this method.
        """
        self._source_collection.reload()

        #
        # Regenerate the frames dataset
        #
        # This assumes that calling `load_view()` when the current patches
        # dataset has been deleted will cause a new one to be generated
        #

        self._frames_dataset.delete()
        _view = self._frames_stage.load_view(self._source_collection)
        self._frames_dataset = _view._frames_dataset

    def _set_labels(self, field_name, sample_ids, label_docs):
        super()._set_labels(field_name, sample_ids, label_docs)

        self._sync_source(fields=[field_name], ids=sample_ids)

    def _delete_labels(self, ids, fields=None):
        super()._delete_labels(ids, fields=fields)

        if fields is not None:
            if etau.is_str(fields):
                fields = [fields]

            frame_fields = [
                self._source_collection._FRAMES_PREFIX + f for f in fields
            ]
        else:
            frame_fields = None

        self._source_collection._delete_labels(ids, fields=frame_fields)

    def _sync_source_sample(self, sample):
        self._sync_source_schema()

        dst_dataset = self._source_collection._root_dataset
        sample_only_fields = self._get_sample_only_fields(
            include_private=True, use_db_fields=True
        )

        updates = {
            k: v
            for k, v in sample.to_mongo_dict().items()
            if k not in sample_only_fields
        }

        if not updates:
            return

        match = {
            "_sample_id": sample._sample_id,
            "frame_number": sample.frame_number,
        }

        dst_dataset._frame_collection.update_one(match, {"$set": updates})

    def _sync_source(self, fields=None, ids=None, update=True, delete=False):
        dst_dataset = self._source_collection._root_dataset
        sample_only_fields = self._get_sample_only_fields(
            include_private=True, use_db_fields=True
        )

        if fields is not None:
            fields = [f for f in fields if f not in sample_only_fields]
            if not fields:
                return

        if update:
            self._sync_source_schema(fields=fields)

            pipeline = []

            if ids is not None:
                pipeline.append(
                    {
                        "$match": {
                            "_id": {"$in": [ObjectId(_id) for _id in ids]}
                        }
                    }
                )

            if fields is None:
                sample_only_fields.discard("_sample_id")
                sample_only_fields.discard("frame_number")

                pipeline.append({"$unset": list(sample_only_fields)})
            else:
                project = {f: True for f in fields}
                project["_id"] = True
                project["_sample_id"] = True
                project["frame_number"] = True
                pipeline.append({"$project": project})

            pipeline.append(
                {
                    "$merge": {
                        "into": dst_dataset._frame_collection_name,
                        "on": ["_sample_id", "frame_number"],
                        "whenMatched": "merge",
                        "whenNotMatched": "discard",
                    }
                }
            )

            self._frames_dataset._aggregate(pipeline=pipeline)

        if delete:
            frame_ids = self._frames_dataset.exclude(self).values("id")
            dst_dataset._clear_frames(frame_ids=frame_ids)

    def _sync_source_schema(self, fields=None, delete=False):
        if delete:
            schema = self.get_field_schema()
        else:
            schema = self._frames_dataset.get_field_schema()

        src_schema = self._source_collection.get_frame_field_schema()
        dst_dataset = self._source_collection._root_dataset

        add_fields = []
        del_fields = []

        if fields is not None:
            # We're syncing specific fields; if they are not present in source
            # collection, add them

            for field_name in fields:
                if field_name not in src_schema:
                    add_fields.append(field_name)
        else:
            # We're syncing all fields; add any missing fields to source
            # collection and, if requested, delete any source fields that
            # aren't in this view

            default_fields = set(
                self._get_default_sample_fields(include_private=True)
            )

            for field_name in schema.keys():
                if (
                    field_name not in src_schema
                    and field_name not in default_fields
                ):
                    add_fields.append(field_name)

            if delete:
                for field_name in src_schema.keys():
                    if field_name not in schema:
                        del_fields.append(field_name)

        for field_name in add_fields:
            field_kwargs = foo.get_field_kwargs(schema[field_name])
            dst_dataset.add_frame_field(field_name, **field_kwargs)

        if delete:
            for field_name in del_fields:
                dst_dataset.delete_frame_field(field_name)

    def _sync_source_keep_fields(self):
        schema = self.get_field_schema()
        src_schema = self._source_collection.get_frame_field_schema()

        del_fields = set(src_schema.keys()) - set(schema.keys())
        if del_fields:
            prefix = self._source_collection._FRAMES_PREFIX
            _del_fields = [prefix + f for f in del_fields]
            self._source_collection.exclude_fields(_del_fields).keep_fields()


def make_frames_dataset(
    sample_collection,
    sample_frames=False,
    fps=None,
    max_fps=None,
    size=None,
    min_size=None,
    max_size=None,
    sparse=False,
    output_dir=None,
    rel_dir=None,
    frames_patt=None,
    force_sample=False,
    skip_failures=True,
    verbose=False,
    name=None,
):
    """Creates a dataset that contains one sample per frame in the video
    collection.

    The returned dataset will contain all frame-level fields and the ``tags``
    of each video as sample-level fields, as well as a ``sample_id`` field that
    records the IDs of the parent sample for each frame.

    By default, ``sample_frames`` is False and this method assumes that the
    frames of the input collection have ``filepath`` fields populated pointing
    to each frame image. Any frames without a ``filepath`` populated will be
    omitted from the frames dataset.

    When ``sample_frames`` is True, this method samples each video in the
    collection into a directory of per-frame images with filenames specified by
    ``frames_patt``. By default, each folder of images is written using the
    same basename as the input video. For example, if
    ``frames_patt = "%%06d.jpg"``, then videos with the following paths::

        /path/to/video1.mp4
        /path/to/video2.mp4
        ...

    would be sampled as follows::

        /path/to/video1/
            000001.jpg
            000002.jpg
            ...
        /path/to/video2/
            000001.jpg
            000002.jpg
            ...

    However, you can use the optional ``output_dir`` and ``rel_dir`` parameters
    to customize the location and shape of the sampled frame folders. For
    example, if ``output_dir = "/tmp"`` and ``rel_dir = "/path/to"``, then
    videos with the following paths::

        /path/to/folderA/video1.mp4
        /path/to/folderA/video2.mp4
        /path/to/folderB/video3.mp4
        ...

    would be sampled as follows::

        /tmp/folderA/
            video1/
                000001.jpg
                000002.jpg
                ...
            video2/
                000001.jpg
                000002.jpg
                ...
        /tmp/folderB/
            video3/
                000001.jpg
                000002.jpg
                ...

    By default, samples will be generated for every video frame at full
    resolution, but this method provides a variety of parameters that can be
    used to customize the sampling behavior.

    .. note::

        If this method is run multiple times with ``sample_frames`` set to
        True, existing frames will not be resampled unless you set
        ``force_sample`` to True.

    .. note::

        The returned dataset is independent from the source collection;
        modifying it will not affect the source collection.

    Args:
        sample_collection: a
            :class:`fiftyone.core.collections.SampleCollection`
        sample_frames (False): whether to assume that the frame images have
            already been sampled at locations stored in the ``filepath`` field
            of each frame (False), or whether to sample the video frames now
            according to the specified parameters (True)
        fps (None): an optional frame rate at which to sample each video's
            frames
        max_fps (None): an optional maximum frame rate at which to sample.
            Videos with frame rate exceeding this value are downsampled
        size (None): an optional ``(width, height)`` at which to sample frames.
            A dimension can be -1, in which case the aspect ratio is preserved.
            Only applicable when ``sample_frames=True``
        min_size (None): an optional minimum ``(width, height)`` for each
            frame. A dimension can be -1 if no constraint should be applied.
            The frames are resized (aspect-preserving) if necessary to meet
            this constraint. Only applicable when ``sample_frames=True``
        max_size (None): an optional maximum ``(width, height)`` for each
            frame. A dimension can be -1 if no constraint should be applied.
            The frames are resized (aspect-preserving) if necessary to meet
            this constraint. Only applicable when ``sample_frames=True``
        sparse (False): whether to only sample frame images for frame numbers
            for which :class:`fiftyone.core.frame.Frame` instances exist in the
            input collection. This parameter has no effect when
            ``sample_frames==False`` since frames must always exist in order to
            have ``filepath`` information used
        output_dir (None): an optional output directory in which to write the
            sampled frames. By default, the frames are written in folders with
            the same basename of each video
        rel_dir (None): a relative directory to remove from the filepath of
            each video, if possible. The path is converted to an absolute path
            (if necessary) via :func:`fiftyone.core.utils.normalize_path`. This
            argument can be used in conjunction with ``output_dir`` to cause
            the sampled frames to be written in a nested directory structure
            within ``output_dir`` matching the shape of the input video's
            folder structure
        frames_patt (None): a pattern specifying the filename/format to use to
            write or check or existing sampled frames, e.g., ``"%%06d.jpg"``.
            The default value is
            ``fiftyone.config.default_sequence_idx + fiftyone.config.default_image_ext``
        force_sample (False): whether to resample videos whose sampled frames
            already exist. Only applicable when ``sample_frames=True``
        skip_failures (True): whether to gracefully continue without raising
            an error if a video cannot be sampled
        verbose (False): whether to log information about the frames that will
            be sampled, if any
        name (None): a name for the dataset

    Returns:
        a :class:`fiftyone.core.dataset.Dataset`
    """
    fova.validate_video_collection(sample_collection)

    if sample_frames != True:
        l = locals()
        for var in ("size", "min_size", "max_size"):
            if l[var]:
                logger.warning(
                    "Ignoring '%s' when sample_frames=%s", var, sample_frames
                )

    if frames_patt is None:
        frames_patt = (
            fo.config.default_sequence_idx + fo.config.default_image_ext
        )

    #
    # Create dataset with proper schema
    #

    dataset = fod.Dataset(name=name, _frames=True)
    dataset.media_type = fom.IMAGE
    dataset.add_sample_field("sample_id", fof.ObjectIdField)

    frame_schema = sample_collection.get_frame_field_schema()
    dataset._sample_doc_cls.merge_field_schema(frame_schema)

    dataset.create_index("sample_id")

    # This index will be used when populating the collection now as well as
    # later when syncing the source collection
    dataset.create_index([("sample_id", 1), ("frame_number", 1)], unique=True)

    _make_pretty_summary(dataset)

    # Initialize frames dataset
    sample_view, frames_to_sample = _init_frames(
        dataset,
        sample_collection,
        sample_frames,
        output_dir,
        rel_dir,
        frames_patt,
        fps,
        max_fps,
        sparse,
        force_sample,
        verbose,
    )

    # Sample frames, if necessary
    if sample_view is not None:
        logger.info("Sampling video frames...")
        fouv.sample_videos(
            sample_view,
            output_dir=output_dir,
            rel_dir=rel_dir,
            frames_patt=frames_patt,
            frames=frames_to_sample,
            size=size,
            min_size=min_size,
            max_size=max_size,
            original_frame_numbers=True,
            force_sample=True,
            save_filepaths=True,
            skip_failures=skip_failures,
        )

    # Merge frame data
    pipeline = []

    if sample_frames == "dynamic":
        pipeline.append({"$unset": "filepath"})

    pipeline.append(
        {
            "$merge": {
                "into": dataset._sample_collection_name,
                "on": ["_sample_id", "frame_number"],
                "whenMatched": "merge",
                "whenNotMatched": "discard",
            }
        }
    )

    sample_collection._aggregate(frames_only=True, post_pipeline=pipeline)

    # Delete samples for frames without filepaths
    if sample_frames == True:
        dataset._sample_collection.delete_many({"filepath": None})

    if sample_frames == False and not dataset:
        logger.warning(
            "Your frames view is empty. Note that you must either "
            "pre-populate the `filepath` field on the frames of your video "
            "collection or pass `sample_frames=True` to this method to "
            "perform the sampling. See "
            "https://voxel51.com/docs/fiftyone/user_guide/using_views.html#frame-views "
            "for more information."
        )

    return dataset


def _make_pretty_summary(dataset):
    set_fields = ["id", "sample_id", "filepath", "frame_number"]
    all_fields = dataset._sample_doc_cls._fields_ordered
    pretty_fields = set_fields + [f for f in all_fields if f not in set_fields]
    dataset._sample_doc_cls._fields_ordered = tuple(pretty_fields)


def _init_frames(
    dataset,
    src_collection,
    sample_frames,
    output_dir,
    rel_dir,
    frames_patt,
    fps,
    max_fps,
    sparse,
    force_sample,
    verbose,
):
    if (
        (sample_frames != False and not sparse)
        or fps is not None
        or max_fps is not None
    ):
        # We'll need frame counts to determine what frames to include/sample
        src_collection.compute_metadata()

    if sample_frames == True and verbose:
        logger.info("Determining frames to sample...")

    #
    # Initialize frames dataset with proper frames
    #

    docs = []
    src_docs = []
    src_inds = []
    missing_filepaths = []

    id_map = {}
    sample_map = defaultdict(set)
    frame_map = defaultdict(set)

    src_dataset = src_collection._root_dataset
    is_clips = src_collection._dataset._is_clips
    if src_collection.has_frame_field("filepath"):
        view = src_collection.select_fields("frames.filepath")
    else:
        view = src_collection.select_fields()

    for sample in view._aggregate(attach_frames=True):
        video_path = sample["filepath"]
        tags = sample.get("tags", [])
        metadata = sample.get("metadata", None) or {}
        frame_rate = metadata.get("frame_rate", None)
        total_frame_count = metadata.get("total_frame_count", -1)
        frames = sample.get("frames", [])

        frame_ids_map = {}
        frames_with_filepaths = set()
        for frame in frames:
            _frame_id = frame["_id"]
            fn = frame["frame_number"]
            filepath = frame.get("filepath", None)

            if sample_frames != False or filepath:
                frame_ids_map[fn] = _frame_id

            if sample_frames == True and filepath:
                frames_with_filepaths.add(fn)

        if is_clips:
            _sample_id = sample["_sample_id"]
            support = sample["support"]
        else:
            _sample_id = sample["_id"]
            support = None

        _outpath = fouv._get_outpath(
            video_path, output_dir=output_dir, rel_dir=rel_dir
        )
        images_patt = os.path.join(os.path.splitext(_outpath)[0], frames_patt)

        # Determine which frame numbers to include in the frames dataset and
        # whether any frame images need to be sampled
        doc_frame_numbers, sample_frame_numbers = _parse_video_frames(
            video_path,
            sample_frames,
            images_patt,
            support,
            total_frame_count,
            frame_rate,
            frame_ids_map,
            force_sample,
            sparse,
            fps,
            max_fps,
            verbose,
        )

        # Record things that need to be sampled
        # Note: [] means no frames, None means all frames
        if sample_frame_numbers != []:
            id_map[video_path] = str(_sample_id)

            if sample_frame_numbers is None:
                sample_map[video_path] = None
            elif sample_map[video_path] is not None:
                sample_map[video_path].update(sample_frame_numbers)

        # Record any already-sampled frames whose `filepath` need to be stored
        # on the source dataset
        if sample_frames == True and sample_frame_numbers is not None:
            missing_fns = (
                set(doc_frame_numbers)
                - set(sample_frame_numbers)
                - frames_with_filepaths
            )
        else:
            missing_fns = set()

        for fn in missing_fns:
            missing_filepaths.append((_sample_id, fn, images_patt % fn))

        # Create necessary frame documents
        for fn in doc_frame_numbers:
            if is_clips:
                fns = frame_map[video_path]
                if fn in fns:
                    continue  # frame has already been sampled

                fns.add(fn)

            _id = frame_ids_map.get(fn, None)

            if sample_frames == "dynamic":
                filepath = video_path
            else:
                filepath = None  # will be populated later

            doc = {
                "filepath": filepath,
                "tags": tags,
                "metadata": None,
                "frame_number": fn,
                "_media_type": "image",
                "_rand": foos._generate_rand(images_patt % fn),
                "_sample_id": _sample_id,
            }

            if _id is not None:
                doc["_id"] = _id
            elif fn in missing_fns:
                # Found a frame whose image is already sampled but for which
                # there is no frame in the source collection. We now need to
                # create a frame so that the missing filepath can be added and
                # the frames dataset can use the same frame ID
                src_docs.append({"_sample_id": _sample_id, "frame_number": fn})
                src_inds.append(len(docs))

            docs.append(doc)

            # Commit batch of docs to frames dataset
            if len(docs) >= 10000:
                _insert_docs(docs, src_docs, src_inds, dataset, src_dataset)

    # Add remaining docs to frames dataset
    _insert_docs(docs, src_docs, src_inds, dataset, src_dataset)

    # Add missing frame filepaths to source collection
    if missing_filepaths:
        logger.info(
            "Setting %d frame filepaths on the input collection that exist "
            "on disk but are not recorded on the dataset",
            len(missing_filepaths),
        )
        src_dataset.add_frame_field("filepath", fof.StringField)
        ops = [
            UpdateOne(
                {"_sample_id": _sample_id, "frame_number": fn},
                {"$set": {"filepath": filepath}},
            )
            for _sample_id, fn, filepath in missing_filepaths
        ]
        src_dataset._bulk_write(ops, frames=True)

    #
    # Finalize which frame images need to be sampled, if any
    #
    # We first populate `sample_map` and then convert to `ids_to_sample` and
    # `frames_to_sample` here to avoid resampling frames when working with clip
    # views with multiple overlapping clips into the same video
    #

    ids_to_sample = []
    frames_to_sample = []
    for video_path, sample_frame_numbers in sample_map.items():
        ids_to_sample.append(id_map[video_path])
        if sample_frame_numbers is not None:
            sample_frame_numbers = sorted(sample_frame_numbers)

        frames_to_sample.append(sample_frame_numbers)

    if ids_to_sample:
        if src_dataset.media_type == fom.GROUP:
            sample_view = src_dataset.select_group_slices(media_type=fom.VIDEO)
        else:
            sample_view = src_dataset

        sample_view = sample_view.select(ids_to_sample, ordered=True)
    else:
        sample_view = None

    return sample_view, frames_to_sample


def _insert_docs(docs, src_docs, src_inds, dataset, src_dataset):
    if src_docs:
        foo.insert_documents(src_docs, src_dataset._frame_collection)

        for idx, src_doc in enumerate(src_docs):
            docs[src_inds[idx]]["_id"] = src_doc["_id"]

        src_docs.clear()
        src_inds.clear()

    if docs:
        foo.insert_documents(docs, dataset._sample_collection)
        docs.clear()


def _parse_video_frames(
    video_path,
    sample_frames,
    images_patt,
    support,
    total_frame_count,
    frame_rate,
    frame_ids_map,
    force_sample,
    sparse,
    fps,
    max_fps,
    verbose,
):
    #
    # Determine target frames, taking subsampling into account
    #

    if fps is not None or max_fps is not None:
        target_frame_numbers = fouv.sample_frames_uniform(
            frame_rate,
            total_frame_count=total_frame_count,
            support=support,
            fps=fps,
            max_fps=max_fps,
        )
    elif support is not None:
        first, last = support
        target_frame_numbers = list(range(first, last + 1))
    else:
        target_frame_numbers = None  # all frames

    #
    # Determine frames for which to generate documents
    #

    if target_frame_numbers is None:
        if total_frame_count < 0:
            doc_frame_numbers = sorted(frame_ids_map.keys())
        else:
            doc_frame_numbers = list(range(1, total_frame_count + 1))
    else:
        doc_frame_numbers = target_frame_numbers

    if sparse or sample_frames == False:
        doc_frame_numbers = [
            fn for fn in doc_frame_numbers if fn in frame_ids_map
        ]

    if sample_frames != True:
        return doc_frame_numbers, []

    #
    # Determine frames that need to be sampled
    #

    if force_sample:
        sample_frame_numbers = doc_frame_numbers
    else:
        sample_frame_numbers = _get_non_existent_frame_numbers(
            images_patt, doc_frame_numbers
        )

    if (
        target_frame_numbers is None
        and len(sample_frame_numbers) == len(doc_frame_numbers)
        and len(doc_frame_numbers) >= total_frame_count
    ):
        sample_frame_numbers = None  # all frames

    if verbose:
        count = total_frame_count if total_frame_count >= 0 else "???"
        if sample_frame_numbers is None:
            logger.info(
                "Must sample all %s frames of '%s'",
                count,
                video_path,
            )
        elif sample_frame_numbers != []:
            logger.info(
                "Must sample %d/%s frames of '%s'",
                len(sample_frame_numbers),
                count,
                video_path,
            )
        else:
            logger.info("Required frames already present for '%s'", video_path)

    return doc_frame_numbers, sample_frame_numbers


def _get_non_existent_frame_numbers(images_patt, frame_numbers):
    return [fn for fn in frame_numbers if not os.path.isfile(images_patt % fn)]



"""
FiftyOne Zoo models provided by :mod:`torchvision:torchvision.models`.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import inspect
from packaging import version

import eta.core.utils as etau

import fiftyone as fo
import fiftyone.core.utils as fou
import fiftyone.utils.torch as fout
import fiftyone.zoo.models as fozm

fou.ensure_torch()
import torchvision


class TorchvisionImageModelConfig(
    fout.TorchImageModelConfig, fozm.HasZooModel
):
    """Configuration for running a :class:`TorchvisionImageModel`.

    Args:
        entrypoint_fcn: a fully-qualified function string like
            ``"torchvision.models.inception_v3"`` specifying the entrypoint
            function that loads the model
        entrypoint_args (None): a dictionary of arguments for
            ``entrypoint_fcn``
        output_processor_cls: a string like
            ``"fifytone.utils.torch.ClassifierOutputProcessor"`` specifying the
            :class:`fifytone.utils.torch.OutputProcessor` to use
        output_processor_args (None): a dictionary of arguments for
            ``output_processor_cls(classes=classes, **kwargs)``
        confidence_thresh (None): an optional confidence threshold apply to any
            applicable predictions generated by the model
        labels_string (None): a comma-separated list of the class names for the
            model, if applicable
        labels_path (None): the path to the labels map for the model, if
            applicable
        mask_targets (None): a mask targets dict for the model, if applicable
        mask_targets_path (None): the path to a mask targets map for the model,
            if applicable
        skeleton (None): a keypoint skeleton dict for the model, if applicable
        image_min_size (None): resize the input images during preprocessing, if
            necessary, so that the image dimensions are at least this
            ``(width, height)``
        image_min_dim (None): resize input images during preprocessing, if
            necessary, so that the smaller image dimension is at least this
            value
        image_max_size (None): resize the input images during preprocessing, if
            necessary, so that the image dimensions are at most this
            ``(width, height)``
        image_max_dim (None): resize input images during preprocessing, if
            necessary, so that the largest image dimension is at most this
            value.
        image_size (None): a ``(width, height)`` to which to resize the input
            images during preprocessing
        image_dim (None): resize the smaller input dimension to this value
            during preprocessing
        image_mean (None): a 3-array of mean values in ``[0, 1]`` for
            preprocessing the input images
        image_std (None): a 3-array of std values in ``[0, 1]`` for
            preprocessing the input images
            inputs that are lists of Tensors
        embeddings_layer (None): the name of a layer whose output to expose as
            embeddings. Prepend ``"<"`` to save the input tensor instead
        use_half_precision (None): whether to use half precision (only
            supported when using GPU)
        cudnn_benchmark (None): a value to use for
            :attr:`torch:torch.backends.cudnn.benchmark` while the model is
            running
    """

    def __init__(self, d):
        d = self.init(d)
        super().__init__(d)


class TorchvisionImageModel(fout.TorchImageModel):
    """Wrapper for evaluating a :mod:`torchvision:torchvision.models` model on
    images.

    Args:
        config: an :class:`TorchvisionImageModelConfig`
    """

    def _download_model(self, config):
        config.download_model_if_necessary()

    def _load_network(self, config):
        entrypoint = etau.get_function(config.entrypoint_fcn)
        kwargs = config.entrypoint_args or {}
        model_dir = fo.config.model_zoo_dir

        monkey_patcher = _make_load_state_dict_from_url_monkey_patcher(
            entrypoint, model_dir
        )
        with monkey_patcher:
            # Builds net and loads state dict from `model_dir`
            model = entrypoint(**kwargs)

        return model


def _make_load_state_dict_from_url_monkey_patcher(entrypoint, model_dir):
    """Monkey patches all instances of ``load_state_dict_from_url()`` that are
    reachable from the given ``entrypoint`` function in the
    :mod:`torchvision:torchvision.models` namespace so that models will be
    loaded from ``model_dir`` and not from the Torch Hub cache directory.
    """
    entrypoint_module = inspect.getmodule(entrypoint)
    if version.parse(torchvision.__version__) >= version.parse("0.12.0"):
        entrypoint_module = torchvision._internally_replaced_utils
    load_state_dict_from_url = entrypoint_module.load_state_dict_from_url

    def custom_load_state_dict_from_url(url, **kwargs):
        return load_state_dict_from_url(url, model_dir=model_dir, **kwargs)

    return fou.MonkeyPatchFunction(
        entrypoint_module,
        custom_load_state_dict_from_url,
        fcn_name=load_state_dict_from_url.__name__,
        namespace=torchvision.models,
    )



"""
FiftyOne v0.8.0 revision.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""


def up(db, dataset_name):
    match_d = {"name": dataset_name}
    dataset_dict = db.datasets.find_one(match_d)

    if "classes" not in dataset_dict:
        dataset_dict["classes"] = {}

    if "default_classes" not in dataset_dict:
        dataset_dict["default_classes"] = []

    if "mask_targets" not in dataset_dict:
        dataset_dict["mask_targets"] = {}

    if "default_mask_targets" not in dataset_dict:
        dataset_dict["default_mask_targets"] = {}

    evaluations = dataset_dict.get("evaluations", {})
    for run_doc in evaluations.values():
        if "results" not in run_doc:
            run_doc["results"] = None

    brain_methods = dataset_dict.get("brain_methods", {})
    for run_doc in brain_methods.values():
        if "results" not in run_doc:
            run_doc["results"] = None

    db.datasets.replace_one(match_d, dataset_dict)


def down(db, dataset_name):
    match_d = {"name": dataset_name}
    dataset_dict = db.datasets.find_one(match_d)

    dataset_dict.pop("classes", None)
    dataset_dict.pop("default_classes", None)

    dataset_dict.pop("mask_targets", None)
    dataset_dict.pop("default_mask_targets", None)

    evaluations = dataset_dict.get("evaluations", {})
    for run_doc in evaluations.values():
        run_doc.pop("results", None)

    brain_methods = dataset_dict.get("brain_methods", {})
    for run_doc in brain_methods.values():
        run_doc.pop("results", None)

    db.datasets.replace_one(match_d, dataset_dict)



"""
Data utilities.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging
import multiprocessing
import requests
import os
import pathlib
import urllib

import eta.core.image as etai
import eta.core.serial as etas
import eta.core.utils as etau
import eta.core.video as etav

import fiftyone.core.fields as fof
import fiftyone.core.labels as fol
import fiftyone.core.utils as fou


logger = logging.getLogger(__name__)


def parse_images_dir(dataset_dir, recursive=True):
    """Parses the contents of the given directory of images.

    Args:
        dataset_dir: the dataset directory
        recursive (True): whether to recursively traverse subdirectories

    Returns:
        a list of image paths
    """
    filepaths = etau.list_files(
        dataset_dir, abs_paths=True, recursive=recursive
    )
    return [p for p in filepaths if etai.is_image_mime_type(p)]


def parse_videos_dir(dataset_dir, recursive=True):
    """Parses the contents of the given directory of videos.

    Args:
        dataset_dir: the dataset directory
        recursive (True): whether to recursively traverse subdirectories

    Returns:
        a list of video paths
    """
    filepaths = etau.list_files(
        dataset_dir, abs_paths=True, recursive=recursive
    )
    return [p for p in filepaths if etav.is_video_mime_type(p)]


def parse_image_classification_dir_tree(dataset_dir):
    """Parses the contents of the given image classification dataset directory
    tree, which should have the following format::

        <dataset_dir>/
            <classA>/
                <image1>.<ext>
                <image2>.<ext>
                ...
            <classB>/
                <image1>.<ext>
                <image2>.<ext>
                ...

    Args:
        dataset_dir: the dataset directory

    Returns:
        samples: a list of ``(image_path, target)`` pairs
        classes: a list of class label strings
    """
    samples = []
    for filepath in etau.list_files(dataset_dir, recursive=True):
        chunks = pathlib.PurePath(filepath).parts

        if any(c.startswith(".") for c in chunks):
            continue

        samples.append((filepath, chunks[0]))

    classes = sorted(set(s[1] for s in samples))
    labels_map_rev = {c: i for i, c in enumerate(classes)}

    samples = [
        (os.path.join(dataset_dir, f), labels_map_rev[c]) for f, c in samples
    ]

    return samples, classes


def download_image_classification_dataset(
    csv_path, dataset_dir, classes=None, num_workers=None
):
    """Downloads the classification dataset specified by the given CSV file,
    which should have the following format::

        <label1>,<image_url1>
        <label2>,<image_url2>
        ...

    The image filenames are the basenames of the URLs, which are assumed to be
    unique.

    The dataset is written to disk in
    :class:`fiftyone.types.FiftyOneImageClassificationDataset` format.

    Args:
        csv_path: a CSV file containing the labels and image URLs
        dataset_dir: the directory to write the dataset
        classes (None): an optional list of classes. By default, this will be
            inferred from the contents of ``csv_path``
        num_workers (None): the number of processes to use to download images.
            By default, ``multiprocessing.cpu_count()`` is used
    """
    labels, image_urls = zip(
        *[
            tuple(line.split(",", 1))
            for line in etau.read_file(csv_path).splitlines()
        ]
    )

    if classes is None:
        classes = sorted(set(labels))

    labels_map_rev = {label: idx for idx, label in enumerate(classes)}

    images_dir = os.path.join(dataset_dir, "data")
    labels_path = os.path.join(dataset_dir, "labels.json")

    logger.info("Downloading images to '%s'...", images_dir)
    image_paths = download_images(
        image_urls, images_dir, num_workers=num_workers
    )

    labels_dict = {}
    for image_path, label in zip(image_paths, labels):
        uuid = os.path.splitext(os.path.basename(image_path))[0]
        labels_dict[uuid] = labels_map_rev[label]

    _labels = {"classes": classes, "labels": labels_dict}

    logger.info("Writing labels to '%s'", labels_path)
    etas.write_json(_labels, labels_path)


def download_images(image_urls, output_dir, num_workers=None):
    """Downloads the images from the given URLs.

    The filenames in ``output_dir`` are the basenames of the URLs, which are
    assumed to be unique.

    Args:
        image_urls: a list of image URLs to download
        output_dir: the directory to write the images
        num_workers (None): the number of processes to use. By default,
            ``multiprocessing.cpu_count()`` is used

    Returns:
        the list of downloaded image paths
    """
    if num_workers is None:
        num_workers = multiprocessing.cpu_count()

    inputs = []
    for url in image_urls:
        filename = os.path.basename(urllib.parse.urlparse(url).path)
        outpath = os.path.join(output_dir, filename)
        inputs.append((url, outpath))

    if num_workers <= 1:
        _download_images(inputs)
    else:
        _download_images_multi(inputs, num_workers)

    return tuple(zip(*inputs))[1]


def _download_images(inputs):
    with fou.ProgressBar() as pb:
        for args in pb(inputs):
            _download_image(args)


def _download_images_multi(inputs, num_workers):
    with fou.ProgressBar(inputs) as pb:
        with fou.get_multiprocessing_context().Pool(
            processes=num_workers
        ) as pool:
            for _ in pb(pool.imap_unordered(_download_image, inputs)):
                pass


def _download_image(args):
    url, outpath = args
    img_bytes = requests.get(url).content
    etau.write_file(img_bytes, outpath)



"""
`Apache Beam <https://beam.apache.org>`_ utilities.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging
import multiprocessing
import numpy as np

import fiftyone.core.dataset as fod
import fiftyone.core.utils as fou
import fiftyone.core.view as fov

fou.ensure_import("apache_beam")

import apache_beam as beam
from apache_beam.options.pipeline_options import PipelineOptions


def beam_import(
    dataset,
    samples,
    parse_fcn=None,
    expand_schema=True,
    validate=True,
    options=None,
    verbose=False,
):
    """Imports the given samples into the dataset via
    `Apache Beam <https://beam.apache.org>`_.

    This function is a parallelized alternative to
    :meth:`fiftyone.core.dataset.Dataset.add_samples`.

    .. note::

        The insertion order of the samples is not guaranteed.

    Example::

        import fiftyone as fo
        import fiftyone.utils.beam as foub

        samples = range(10000)

        def make_sample(idx):
            return fo.Sample(filepath="image%d.png" % idx, uuid=idx)

        #
        # Option 1: build the samples on the workers
        #

        dataset = fo.Dataset()

        foub.beam_import(dataset, samples, parse_fcn=make_sample)
        print(dataset)

        #
        # Option 2: build the samples in the main thread
        #
        # This is generally not preferred but may be necessary if your
        # ``parse_fcn`` is not serializable
        #

        dataset = fo.Dataset()

        samples = map(make_sample, samples)

        foub.beam_import(dataset, samples)
        print(dataset)

    Args:
        dataset: a :class:`fiftyone.core.dataset.Dataset`
        samples: an iterable of samples. If no ``parse_fcn`` is provided, these
            must be :class:`fiftyone.core.sample.Sample` instances. If a
            ``parse_fcn`` is provided, these are passed to it for parsing
        parse_fcn (None): an optional function that converts elements of
            ``samples`` to :class:`fiftyone.core.sample.Sample` instances
        expand_schema (True): whether to dynamically add new sample fields
            encountered to the dataset schema. If False, an error is raised
            if a sample's schema is not a subset of the dataset schema
        validate (True): whether to validate that the fields of each sample
            are compliant with the dataset schema before adding it
        options (None): a
            ``apache_beam.options.pipeline_options.PipelineOptions`` that
            configures how to run the pipeline. By default, the pipeline will
            be run via Beam's direct runner using
            ``multiprocessing.cpu_count()`` threads
        verbose (False): whether to log the Beam pipeline's messages
    """
    if options is None:
        options = PipelineOptions(
            runner="direct",
            direct_num_workers=multiprocessing.cpu_count(),
            direct_running_mode="multi_threading",
        )

    sample0, samples = _pop_first(samples)

    if sample0 is None:
        return  # empty

    if parse_fcn is not None:
        sample0 = parse_fcn(sample0)

    # Manually insert first sample to reduce chances of parallel schema changes
    dataset.add_sample(sample0, expand_schema=expand_schema, validate=validate)

    if parse_fcn is None:
        # `Sample` objects are not serializable so we must manually serialize
        # and deserialize in `ImportBatch`
        samples = map(lambda s: s.to_mongo_dict(include_id=True), samples)

    import_batch = ImportBatch(
        dataset.name,
        parse_fcn=parse_fcn,
        expand_schema=expand_schema,
        validate=validate,
    )

    logger = logging.getLogger()
    level = logger.level if verbose else logging.CRITICAL
    with fou.SetAttributes(logger, level=level):
        with beam.Pipeline(options=options) as pipeline:
            _ = (
                pipeline
                | "InitImport" >> beam.Create(samples)
                | "ImportBatches" >> beam.ParDo(import_batch)
            )

    # The dataset's schema may have changed in another process
    dataset.reload()


def beam_merge(
    dataset, samples, parse_fcn=None, options=None, verbose=False, **kwargs
):
    """Merges the given samples into the dataset via
    `Apache Beam <https://beam.apache.org>`_.

    This function is a parallelized alternative to
    :meth:`fiftyone.core.dataset.Dataset.merge_samples`.

    .. note::

        This function is only useful for merging **in-memory samples** into a
        dataset. If you are merging a sample collection, simply call
        :meth:`fiftyone.core.dataset.Dataset.merge_samples`.

    Example::

        import fiftyone as fo
        import fiftyone.utils.beam as foub
        import fiftyone.zoo as foz

        dataset = foz.load_zoo_dataset("quickstart").clone()

        samples = iter(dataset.select_fields("predictions"))

        foub.beam_merge(dataset, samples, fields={"predictions": "predictions2"})

        print(dataset.count("predictions.detections"))
        print(dataset.count("predictions2.detections"))

    Args:
        dataset: a :class:`fiftyone.core.dataset.Dataset`
        samples: an iterable of samples. If no ``parse_fcn`` is provided, these
            must be :class:`fiftyone.core.sample.Sample` instances. If a
            ``parse_fcn`` is provided, these are passed to it for parsing
        parse_fcn (None): an optional function that converts elements of
            ``samples`` to :class:`fiftyone.core.sample.Sample` instances
        options (None): a
            ``apache_beam.options.pipeline_options.PipelineOptions`` that
            configures how to run the pipeline. By default, the pipeline will
            be run via Beam's direct runner using
            ``multiprocessing.cpu_count()`` threads
        verbose (False): whether to log the Beam pipeline's messages
        **kwargs: keyword arguments for
            :meth:`fiftyone.core.dataset.Dataset.merge_samples`
    """

    # If the merge does not require a `key_fcn`, then it is fastest to import
    # the samples into a temporary collection and then merge that
    if kwargs.get("key_fcn", None) is None:
        tmp_dataset = fod.Dataset()

        try:
            beam_import(
                tmp_dataset,
                samples,
                parse_fcn=parse_fcn,
                options=options,
                verbose=verbose,
            )

            dataset.merge_samples(tmp_dataset, **kwargs)
        finally:
            tmp_dataset.delete()

        return

    if options is None:
        options = PipelineOptions(
            runner="direct",
            direct_num_workers=multiprocessing.cpu_count(),
            direct_running_mode="multi_threading",
        )

    if parse_fcn is None:
        # `Sample` objects are not serializable so we must manually serialize
        # and deserialize in `MergeBatch`
        samples = map(lambda s: s.to_mongo_dict(include_id=True), samples)

    merge_batch = MergeBatch(dataset.name, parse_fcn=parse_fcn, **kwargs)

    logger = logging.getLogger()
    level = logger.level if verbose else logging.CRITICAL
    with fou.SetAttributes(logger, level=level):
        with beam.Pipeline(options=options) as pipeline:
            _ = (
                pipeline
                | "InitMerge" >> beam.Create(samples)
                | "MergeBatches" >> beam.ParDo(merge_batch)
            )

    # The dataset's schema may have changed in another process
    dataset.reload()


def beam_export(
    sample_collection,
    num_shards,
    options=None,
    verbose=False,
    render_kwargs=None,
    **kwargs,
):
    """Exports the given sample collection in the specified number shards via
    `Apache Beam <https://beam.apache.org>`_.

    This function is a parallelized alternative to
    :meth:`fiftyone.core.collections.SampleCollection.export` that effectively
    performs the following sharded export in parallel::

        for idx, (first, last) in enumerate(shards, 1):
            _kwargs = render_kwargs(kwargs, idx)
            sample_collection[first:last].export(**_kwargs)

    Example::

        from apache_beam.options.pipeline_options import PipelineOptions

        import fiftyone as fo
        import fiftyone.utils.beam as foub
        import fiftyone.zoo as foz

        dataset = foz.load_zoo_dataset("quickstart")

        # Use multithreading instead of the default multiprocessing
        options = PipelineOptions(
            runner="direct",
            direct_num_workers=10,
            direct_running_mode="multi_threading",
        )

        foub.beam_export(
            dataset,
            num_shards=20,
            options=options,
            dataset_type=fo.types.TFObjectDetectionDataset,
            label_field="ground_truth",
            tf_records_path="/tmp/beam/tf.records-%05d-of-00020",
        )

    Args:
        sample_collection: a
            :class:`fiftyone.core.collections.SampleCollection`
        num_shards: the number of shards to write
        options (None): a
            ``apache_beam.options.pipeline_options.PipelineOptions`` that
            configures how to run the pipeline. By default, the pipeline will
            be run via Beam's direct runner using
            ``min(num_shards, multiprocessing.cpu_count())`` processes
        verbose (False): whether to log the Beam pipeline's messages
        render_kwargs (None): a function that renders ``kwargs`` for the
            current shard. The function should have signature
            ``def render_kwargs(kwargs, idx) -> kwargs``, where ``idx`` in
            ``[1, num_shards]`` is the shard index. By default, any
            string-valued arguments that contain format patterns like ``%05d``
            will be rendered via ``value % idx``
        **kwargs: keyword arguments for
            :meth:`fiftyone.core.collections.SampleCollection.export`
    """
    if options is None:
        num_workers = min(num_shards, multiprocessing.cpu_count())
        options = PipelineOptions(
            runner="direct",
            direct_num_workers=num_workers,
            direct_running_mode="multi_processing",
        )

    if isinstance(sample_collection, fov.DatasetView):
        dataset_name = sample_collection._root_dataset.name
        view_stages = sample_collection._serialize()
    else:
        dataset_name = sample_collection.name
        view_stages = None

    export_batch = ExportBatch(
        dataset_name,
        view_stages=view_stages,
        render_kwargs=render_kwargs,
    )

    n = len(sample_collection)
    edges = [int(round(b)) for b in np.linspace(0, n, num_shards + 1)]

    batches = [
        {"idx": idx, "start": start, "stop": stop}
        for idx, (start, stop) in enumerate(zip(edges[:-1], edges[1:]), 1)
    ]

    logger = logging.getLogger()
    level = logger.level if verbose else logging.CRITICAL
    with fou.SetAttributes(logger, level=level):
        with beam.Pipeline(options=options) as pipeline:
            _ = (
                pipeline
                | "InitExport" >> beam.Create(batches)
                | "ExportBatches" >> beam.ParDo(export_batch, **kwargs)
            )


class ImportBatch(beam.DoFn):
    def __init__(
        self, dataset_name, parse_fcn=None, expand_schema=True, validate=True
    ):
        self.dataset_name = dataset_name
        self.parse_fcn = parse_fcn
        self.expand_schema = expand_schema
        self.validate = validate

        self._dataset = None
        self._samples = None

    def setup(self):
        import fiftyone as fo

        self._dataset = fo.load_dataset(self.dataset_name)

    def start_bundle(self):
        self._samples = []

    def process(self, sample):
        import fiftyone as fo

        if self.parse_fcn is not None:
            sample = self.parse_fcn(sample)
        else:
            sample = fo.Sample.from_dict(sample)

        self._samples.append(sample)

    def finish_bundle(self):
        if self._samples:
            self._dataset._add_samples_batch(
                self._samples, self.expand_schema, self.validate
            )


class MergeBatch(beam.DoFn):
    def __init__(
        self,
        dataset_name,
        parse_fcn=None,
        key_field="filepath",
        key_fcn=None,
        expand_schema=True,
        **kwargs,
    ):
        self.dataset_name = dataset_name
        self.parse_fcn = parse_fcn
        self.key_field = key_field
        self.key_fcn = key_fcn
        self.expand_schema = expand_schema
        self.kwargs = kwargs

        self._dataset = None
        self._id_map = None
        self._key_fcn = None
        self._samples = None

    def setup(self):
        import fiftyone as fo

        dataset = fo.load_dataset(self.dataset_name)

        if self.key_fcn is None:
            key_field = self.key_field
            key_fcn = lambda sample: sample[key_field]
            id_map = {k: v for k, v in zip(*dataset.values([key_field, "id"]))}
        else:
            key_fcn = self.key_fcn
            id_map = {key_fcn(s): s.id for s in dataset}

        self._dataset = dataset
        self._key_fcn = key_fcn
        self._id_map = id_map

    def start_bundle(self):
        self._samples = []

    def process(self, sample):
        import fiftyone as fo

        if self.parse_fcn is not None:
            sample = self.parse_fcn(sample)
        else:
            sample = fo.Sample.from_dict(sample)

        self._samples.append(sample)

    def finish_bundle(self):
        import fiftyone.core.dataset as fod
        import fiftyone.core.utils as fou

        if not self._samples:
            return

        kwargs, _ = fou.extract_kwargs_for_function(
            fod._make_merge_samples_generator, self.kwargs
        )

        samples = fod._make_merge_samples_generator(
            self._dataset,
            self._samples,
            self._key_fcn,
            self._id_map,
            expand_schema=self.expand_schema,
            **kwargs,
        )

        self._dataset._upsert_samples_batch(
            list(samples), self.expand_schema, True
        )


class ExportBatch(beam.DoFn):
    def __init__(self, dataset_name, view_stages=None, render_kwargs=None):
        self.dataset_name = dataset_name
        self.view_stages = view_stages
        self.render_kwargs = render_kwargs or self.default_render_kwargs

        self._sample_collection = None

    @staticmethod
    def default_render_kwargs(kwargs, idx):
        _kwargs = {}
        for k, v in kwargs.items():
            if isinstance(v, str):
                try:
                    _kwargs[k] = v % idx
                except:
                    _kwargs[k] = v
            else:
                _kwargs[k] = v

        return _kwargs

    def setup(self):
        import fiftyone as fo
        import fiftyone.core.view as fov

        dataset = fo.load_dataset(self.dataset_name)

        if self.view_stages:
            sample_collection = fov.DatasetView._build(
                dataset, self.view_stages
            )
        else:
            sample_collection = dataset

        self._sample_collection = sample_collection

    def process(self, element, **kwargs):
        import fiftyone as fo
        import fiftyone.core.utils as fou

        idx = element["idx"]
        start = element["start"]
        stop = element["stop"]
        kwargs = self.render_kwargs(kwargs, idx)

        with fou.SetAttributes(fo.config, show_progress_bars=False):
            self._sample_collection[start:stop].export(**kwargs)


def _pop_first(x):
    xi = iter(x)

    try:
        x0 = next(xi)
    except StopIteration:
        x0 = None

    return x0, xi
